import os
import sys
import json
import shutil
import numpy as np
from lxml import etree
from pymatgen.io.vasp.outputs import Vasprun


def extract_from_lammps(species: str,
                        lammps: str,
                        prefix_pos: str,
                        prefix_force: str,
                        prefix_cond: str):
    """
    Will be updated.
    """
    pass

def extract_from_vasp(species: str,
                      vasprun: str = "vasprun.xml",
                      prefix_pos: str = "pos",
                      prefix_force: str = "force",
                      prefix_cond: str = "cond"):
    """
    Extracting VacHopPy input files from vasprun.xml.
    Atomic trajectories are unwrapped according to PBC condition.

    Args:
        species (str): atom species (e.g., "O").
        vasprun (str, optional): vasprun.xml file. Defaults to "vasprun.xml".
        prefix_pos (str, optional): prefix for pos file. Defaults to "pos".
        prefix_force (str, optional): prefix for force file. Defaults to "force".
        prefix_cond (str, optional): prefix for cond file. Defaults to "cond".
    """

    if not os.path.isfile(vasprun):
        print(f"{vasprun} is not found.")
        sys.exit(0)

    v = Vasprun(vasprun, 
                parse_dos=False, 
                parse_eigen=False,
                parse_potcar_file=False)
    structure = v.final_structure
    atom_symbols = [str(site.specie) for site in structure.sites]

    # Atom count dictionary
    from collections import Counter
    atom_counts = dict(Counter(atom_symbols))

    # Target indices for selected species
    target_indices = [i for i, sym in enumerate(atom_symbols) if sym == species]
    if not target_indices:
        raise ValueError(f"No atoms with symbol '{species}' found.")

    iterations = v.ionic_steps
    nsw = len(iterations)
    n_atoms = len(target_indices)

    pos = np.zeros((nsw, n_atoms, 3), dtype=np.float64)
    force = np.zeros((nsw, n_atoms, 3), dtype=np.float64)

    for step_idx, step in enumerate(iterations):
        for j, atom_idx in enumerate(target_indices):
            pos[step_idx, j, :] = step["structure"].sites[atom_idx].frac_coords
            force[step_idx, j, :] = step["forces"][atom_idx]

    # PBC refinement
    displacement = np.zeros_like(pos)
    displacement[0:] = 0
    displacement[1:, :] = np.diff(pos, axis=0)
    displacement[displacement>0.5] -= 1.0
    displacement[displacement<-0.5] += 1.0
    displacement = np.cumsum(displacement, axis=0)
    pos = pos[0] + displacement
    
    # save positions and forces
    np.save(f"{prefix_pos}.npy", pos)
    np.save(f"{prefix_force}.npy", force)
    print(f"{prefix_pos}.npy is created.")
    print(f"{prefix_force}.npy is created.")

    # Extract metadata
    incar = v.incar
    potim = float(incar.get("POTIM", -1))
    nblock = int(incar.get("NBLOCK", -1))
    tebeg = float(incar.get("TEBEG", -1))
    teend = float(incar.get("TEEND", -1))

    if tebeg != teend:
        raise ValueError(f"TEBEG ({tebeg}) and TEEND ({teend}) are not equal.")

    lattice = structure.lattice.matrix.tolist()  # 3x3 list

    cond = {
        "symbol": species,
        "nsw": nsw,
        "potim": potim,
        "nblock": 1, # for vasprun.xml, all steps are stored
        "temperature": tebeg,
        "atom_counts": atom_counts,
        "lattice": lattice
    }
    
    cond_file = f"{prefix_cond}.json"
    with open(cond_file, "w") as f:
        json.dump(cond, f, indent=2)

    print(f"{cond_file} is created.")


def combine_vasprun(vasprun1, 
                    vasprun2,
                    vasprun_out = "vasprun_combined.xml"):
    # Load and merge files
    v1 = etree.parse(vasprun1)
    v2 = etree.parse(vasprun2)

    r1 = v1.getroot()
    r2 = v2.getroot()

    # remove duplicated <parameters> from v2
    p2 = r2.find("parameters")
    if p2 is not None:
        r2.remove(p2)

    # append calculations from v2
    calcs2 = r2.findall("calculation")
    for c in calcs2:
        r1.append(c)

    # recalculate total steps
    nsw_total = len(r1.findall("calculation"))

    # fix ALL <i name="NSW"> across entire tree
    for elem in r1.findall(".//i[@name='NSW']"):
        elem.text = str(nsw_total)

    # save result
    etree.ElementTree(r1).write(vasprun_out, pretty_print=True)
    

def crop_vasprun(vasprun,
                 nsw_crop,
                 vasprun_out="vasprun_cropped.xml"):
    print(f"Cropping first {nsw_crop} iterations from {vasprun}...")
    
    tree = etree.parse(vasprun)
    root = tree.getroot()

    calculations = root.findall(".//calculation")
    total = len(calculations)
    for calc in calculations[nsw_crop:]:
        calc.getparent().remove(calc)

    print(f"  Original steps: {total}, Cropped to: {nsw_crop}")

    # fix ALL <i name="NSW"> across entire tree
    for elem in root.findall(".//i[@name='NSW']"):
        elem.text = str(nsw_crop)

    tree.write(vasprun_out, pretty_print=True, encoding='utf-8', xml_declaration=True)
    print(f"Saved cropped file to {vasprun_out}")
 

def CosineDistance(fp1, fp2):
    """
    fp1 : fingerprint of structure 1
    fp2 : fingerprint of structure 2
    """
    dot = np.dot(fp1, fp2)
    norm1 = np.linalg.norm(fp1, ord=2)
    norm2 = np.linalg.norm(fp2, ord=2)

    return 0.5 * (1 - dot/(norm1*norm2))



import os
import sys
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from colorama import Fore   # color map for tqdm

BOLD = '\033[1m'
CYAN = '\033[36m'
MAGENTA = '\033[35m'
GREEN = '\033[92m' # Green color
RED = '\033[91m'   # Red color
RESET = '\033[0m'  # Reset to default color

class EinsteinRelation:
    def __init__(self, 
                 xdatcar="",
                 outcar="",
                 segments=1, 
                 skip=0,
                 skip2=0, 
                 verbose=True,
                 getD=True):
        
        if os.path.isfile(xdatcar):
            self.xdatcar = xdatcar
        else:
            print(f"{xdatcar} is not found.")
            sys.exit(0)

        if os.path.isfile(outcar):
            self.outcar = outcar
        else:
            print(f"{outcar} is not found.")
            sys.exit()
            
        self.verbose = verbose
        
        # read outcar
        self.potim = None
        self.nblock = None
        self.read_outcar()

        # read XDATCAR
        self.TypeName = None    # List of atom species
        self.Ntype = None       # Number of atom species
        self.Nions = None       # Total number of atom
        self.Nelem = None       # Number of each atoms
        self.Niter = None       # Number of steps
        self.position = None    # Direct coordination of atoms
        self.cell = None        # Lattice of cell
        self.read_xdatcar()

        self.ChemSymb = None
        self.list_symbol()

        # Parameters for skip
        self.segments = segments    # Dividing whole MD steps into segments
        self.skip = skip            # Skipping initial steps for whole MD steps
        self.skip2 = skip2          # Skipping initial steps for each segments

        Time = self.Niter * self.potim * self.nblock
        if self.verbose:
            print("\t--- MD information ---")
            print(f"\ttotal time = {Time/1000} ps")

        # Mean square distance
        self.msd = None
        self.get_msd()
        
        # Diffusivity.
        # For ensemble calculation, please set getD = False to save time
        self.diffcoeff = None
        self.intercept = None
        self.ddiffcoeff = None  # Deviation of D

        if getD is True:
            self.get_diffcoeff()

        if self.verbose:
            self.plot_msd()
    
    
    def read_outcar(self):
        if self.verbose:
            print(f"reading {self.outcar}...")
        
        check_potim, check_nblock = False, False
        with open(self.outcar, 'r') as f:
            for line in f:
                if 'POTIM ' in line:
                    self.potim = float(line.split()[2])
                    check_potim = True
                if 'NBLOCK' in line:
                    self.nblock = int(line.split()[2].replace(';',''))
                    check_nblock = True
                if check_potim and check_nblock:
                    break

        if self.verbose:
            print(f"\tpotim = {self.potim}")
            print(f"\tnblock = {self.nblock}\n")
    
    
    def read_xdatcar(self):
        if self.verbose:
            print(f"reading {self.xdatcar}...")
        with open(self.xdatcar,'r') as file:
            inp = [line.strip() for line in file]
        scale = float(inp[1])
        self.cell = np.array([line.split() for line in inp[2:5]], dtype=float)
        self.cell *= scale

        ta = inp[5].split()
        tb = inp[6].split()
        if ta[0].isalpha():
            self.TypeName = ta
            self.Ntype = len(ta)
            self.Nelem = np.array(tb, dtype=int)
            self.Nions = self.Nelem.sum()
        if self.verbose:
            print(f"\tatom type = {self.TypeName}")
            print(f"\tnumber of type = {self.Ntype}")
            print(f"\tnumber of atom = {self.Nelem}")
            print(f"\ttotal number of atom = {self.Nions}")

        pos = np.array(
            [line.split() for line in inp[7:] if not line.split()[0].isalpha()],
            dtype=float)
        self.position = pos.flatten().reshape((-1,self.Nions,3))
        self.Niter = self.position.shape[0]
        if self.verbose:
            print(f"\tshape of position = {self.position.shape} #(nsw, number of atom, xyz)\n")


    def list_symbol(self):
        self.ChemSymb = []
        for i in range(self.Ntype):
            self.ChemSymb += [np.tile(self.TypeName[i], self.Nelem[i])]
        self.ChemSymb = np.concatenate(self.ChemSymb)
    
    
    def get_msd(self):
        # length of segment
        seglength = int(np.floor((self.Niter-self.skip)/self.segments))
        if self.verbose:
            print(f"\tnumber of segments = {self.segments}")
            print(f"\tsegment length = {seglength}")

        # msd for (each segment), (each type of element), (each step), and, (x, y, z)
        self.msd=np.zeros(shape=(self.segments, self.Ntype, seglength, 3))

        for i in range(self.segments):
            # displacement [step,atom_idx,xyz]
            displacement = np.zeros_like(
                self.position[self.skip+seglength*i:self.skip+seglength*(i+1),:])
            displacement[0,:,:] = 0
            displacement[1:,:,:] = np.diff(
                self.position[self.skip+seglength*i:self.skip+seglength*(i+1),:], axis=0)

            # wrap back into box
            displacement[displacement > 0.5] -= 1.0
            displacement[displacement < -0.5] += 1.0

            # vector of accumulated displacement
            displacement = np.cumsum(displacement, axis=0)
        
            displacementC = np.zeros_like(displacement)
            displacementC = np.dot(displacement[:,:,:], self.cell)  # Direct to Cartesian

            # squared displacement in x, y, z
            displacementC[:,:,:] =  displacementC[:,:,:]**2

            for j in range(self.Ntype):
                if j==0:
                    labelst=0
                else:
                    labelst = np.sum(self.Nelem[:j])
                labeled = np.sum(self.Nelem[:j+1])
                self.msd[i,j,:,:] = np.average(displacementC[:,labelst:labeled,:], axis=1) 

        if self.verbose:
            print(f"\tshape of msd = {self.msd.shape} #(segment, number of type, nsw - skip, xyz)\n")
                
                
    def get_diffcoeff(self):
        self.diffcoeff = np.zeros(shape=(self.segments, self.Ntype, 3))
        self.intercept = np.zeros(shape=(self.segments, self.Ntype, 3))
        self.ddiffcoeff = np.zeros(shape=(self.Ntype, 3))

        for i in range(self.segments):           # loop over segments
            for j in range(self.Ntype):          # loop over type of elements
                for k in range(3):               # loop over x.y,z
                    # time unit in fs, distance unit in A, unit in m^2/s
                    if np.floor((self.Niter-self.skip)/self.segments) < 2*self.skip2:   
                        # too large skip2... skip2 is automatically set to 0
                        self.diffcoeff[i,j,k], self.intercept[i,j,k]=np.polyfit(
                                self.potim*self.nblock*
                                np.arange(np.floor((self.Niter-self.skip)/self.segments)),
                                1E-5*self.msd[i,j,:,k], 
                                deg=1)
                    else:
                        self.diffcoeff[i,j,k], self.intercept[i,j,k] = np.polyfit(
                                self.potim*self.nblock*
                                np.arange(np.floor((self.Niter-self.skip)/self.segments)-self.skip2), 
                                1E-5*self.msd[i,j,self.skip2:,k],
                                deg=1)
                        
        # standard deviation of D over segments for x,y,z directions
        self.diffcoeff /= 2      # x, y, z 각각에 대해 계산되었음 -> 1차원임 -> 2로 나누어짐
        self.ddiffcoeff = np.std(self.diffcoeff, axis=0) / 2
    
    
    def get_msd_atom(self, 
                     symbol):
        if symbol in self.TypeName:
            idx = np.where(np.array(self.TypeName) == symbol)[0][0]
            return self.msd[:,idx,:,:]  # shape: [1, steps, 3:xyz]
        else:
            print("No matched atom!")
            return None
        
        
    def plot_msd(self):
        time = np.arange(self.skip, self.Niter) * self.potim / 1000
        for type in self.TypeName:
            msd = np.sum(self.get_msd_atom(symbol=type).squeeze(), axis=1)
            plt.plot(time, msd, label=type)
        plt.xlabel('t (ps)', fontsize=13)
        plt.ylabel(r'MSD ($Å^2$)', fontsize=13)
        # plt.ylabel(r'$<\Delta r^{2}> (Å^2)$', fontsize=13)
        plt.legend(fontsize=13)
        plt.show()



class EnsembleEinstein:
    def __init__(self, 
                 symbol, 
                 prefix, 
                 labels, 
                 segments, 
                 skip, 
                 start=None, 
                 end=None):
        ## XDATCAR format : XDATCAR_{label}
        ## OUTCAR format : OUTCAR

        ## Arguments
        self.symbol = symbol        # ex: 'O'
        self.prefix = prefix        # ex: 'xdatcar/xdatcar.2000K/'
        if self.prefix[-1] != '/':
            self.prefix += '/'
        self.labels = labels        # ex: np.arange(1000, 20000+1, 1000)
        self.segments = segments    # See MD2D for description about segments
        self.skip = skip
        ## args : 'start' and 'end' : Fitting range

        # Properties of NVT ensembles
        self.msd_xyz = None         # shape: (steps, 3;xyz)
        self.msd = None             # shape: (steps, )
        self.msd_x = None
        self.msd_y = None
        self.msd_z = None

        self.potim = None
        self.nblock = None

        self.diffcoeff = None       # m2/s
        self.intercept = None
        
        self.diffcoeff_x = None     
        self.intercept_x = None
        
        self.diffcoeff_y = None     
        self.intercept_y = None
        
        self.diffcoeff_z = None     
        self.intercept_z = None

        # For linear fitting
        self.start = start          # Step number in MD simulation
        self.end = end
        self.timestep = None        # [ps]

        self.get_ensemble_msd()

        if start is not None and end is not None:
            _, __ = self.get_ensembleD()
            _, __ = self.get_ensembleD_x()
            _, __ = self.get_ensembleD_y()
            _, __ = self.get_ensembleD_z()


    def get_ensemble_msd(self):
        desc = self.prefix.split('/')[-1] if len(self.prefix.split('/')[-1]) > 0 \
            else  self.prefix.split('/')[-2]
        desc = desc.split('.')[-1]
        for label in tqdm(self.labels,
                          bar_format='{l_bar}{bar:20}{r_bar}{bar:-10b}',
                          ascii=True,
                          desc=f'{RED}{BOLD}{desc:>9s}{RESET}'):
            xdatcar = self.prefix + "XDATCAR_" + str(label)
            outcar = self.prefix + "OUTCAR"
            ensemble = EinsteinRelation(xdatcar=xdatcar,
                                        outcar=outcar,
                                        segments=self.segments,
                                        skip=self.skip,
                                        verbose=False,
                                        getD=False)
            if self.msd_xyz is None:
                self.msd_xyz = ensemble.get_msd_atom(symbol=self.symbol)
                self.potim = ensemble.potim
                self.nblock = ensemble.nblock
            else:
                self.msd_xyz = np.append(self.msd_xyz, 
                                         ensemble.get_msd_atom(symbol=self.symbol), axis=0)
        
        self.msd_xyz = np.mean(self.msd_xyz, axis=0)
        self.msd = np.sum(self.msd_xyz, axis=1)
        self.timestep = np.arange(self.msd[:].shape[0]) * self.potim * self.nblock / 1000


    def get_ensembleD(self, 
                      *args):
        if len(args) == 2:
            self.start, self.end = args[0], args[1]
        if self.start is None or self.end is None:
            sys.exit("start and end are needed to determine fitting range.")

        self.diffcoeff, self.intercept = np.polyfit(6*self.potim*self.nblock*np.arange(self.end-self.start),
                                                    1E-5*self.msd[self.start:self.end], 
                                                    deg=1)
        return self.diffcoeff, self.intercept
    
    def plot_ensemble_msd(self):
        temp = self.prefix.split('/')[-2].split('.')[-1]
        plt.plot(self.timestep, self.msd[:], label=f"{temp}")

        if self.diffcoeff is not None:
            # visualize fitting line
            plt.plot(np.arange(self.start, self.end)*self.potim*self.nblock/1000, 
                     self.diffcoeff*(6*self.potim*self.nblock*1.0E5)*
                     np.arange(self.end-self.start)+self.intercept*1.0E5, 
                     'k:')
            # visualize fitting range
            plt.axvline(self.start*self.potim*self.nblock/1000, 0, 1, color='black', linewidth=0.5)
            plt.axvline(self.end*self.potim*self.nblock/1000, 0, 1, color='black', linewidth=0.5)

        plt.legend(loc='upper left')
        plt.xlabel('t (ps)', fontsize=13)
        plt.ylabel(r'MSD ($Å^2$)', fontsize=13)


    def get_ensembleD_x(self, 
                        *args):
        if len(args) == 2:
            self.start, self.end = args[0], args[1]
        if self.start is None or self.end is None:
            sys.exit("start and end are needed to determine fitting range.")

        self.msd_x = self.msd_xyz[:,0]
        self.diffcoeff_x, self.intercept_x = np.polyfit(2*self.potim*self.nblock*
                                                        np.arange(self.end-self.start),
                                                        1E-5*self.msd_x[self.start:self.end],
                                                        deg=1)
        return self.diffcoeff_x, self.intercept_x
    
    
    def plot_ensemble_msd_x(self):
        temp = self.prefix.split('/')[-2].split('.')[-1]
        plt.plot(self.timestep, self.msd_x[:], label=f"{temp}")

        if self.diffcoeff_x is not None:
            # visualize fitting line
            plt.plot(np.arange(self.start, self.end) * self.potim * self.nblock / 1000, 
                     self.diffcoeff_x*(2*self.potim*self.nblock*1.0E5)*
                     np.arange(self.end-self.start)+self.intercept_x*1.0E5, 
                     'k:')
            # visualize fitting range
            plt.axvline(self.start* self.potim * self.nblock / 1000, 0, 1, color='black', linewidth=0.5)
            plt.axvline(self.end* self.potim * self.nblock / 1000, 0, 1, color='black', linewidth=0.5)

        plt.title("x component", fontsize=14)
        plt.legend(loc='upper left')
        plt.xlabel('t (ps)', fontsize=13)
        plt.ylabel(r'MSD ($Å^2$)', fontsize=13)


    def get_ensembleD_y(self, 
                        *args):
        if len(args) == 2:
            self.start, self.end = args[0], args[1]
        if self.start is None or self.end is None:
            sys.exit("start and end are needed to determine fitting range.")

        self.msd_y = self.msd_xyz[:,1]
        self.diffcoeff_y, self.intercept_y = np.polyfit(2*self.potim*self.nblock*
                                                        np.arange(self.end - self.start), 
                                                        1E-5*self.msd_y[self.start:self.end],
                                                        deg=1)
        return self.diffcoeff_y, self.intercept_y
    
    
    def plot_ensemble_msd_y(self):
        temp = self.prefix.split('/')[-2].split('.')[-1]
        plt.plot(self.timestep, self.msd_y[:], label=f"{temp}")

        if self.diffcoeff_y is not None:
            # visualize fitting line
            plt.plot(np.arange(self.start, self.end) * self.potim * self.nblock / 1000, 
                     self.diffcoeff_y*(2*self.potim*self.nblock*1.0E5)*
                     np.arange(self.end-self.start)+self.intercept_y*1.0E5, 
                     'k:')
            # visualize fitting range
            plt.axvline(self.start* self.potim * self.nblock / 1000, 0, 1, color='black', linewidth=0.5)
            plt.axvline(self.end* self.potim * self.nblock / 1000, 0, 1, color='black', linewidth=0.5)

        plt.title("y component", fontsize=14)
        plt.legend(loc='upper left')
        plt.xlabel('t (ps)', fontsize=13)
        plt.ylabel(r'MSD ($Å^2$)', fontsize=13)


    def get_ensembleD_z(self, 
                        *args):
        if len(args) == 2:
            self.start, self.end = args[0], args[1]
        if self.start is None or self.end is None:
            sys.exit("start and end are needed to determine fitting range.")

        self.msd_z = self.msd_xyz[:,2]
        self.diffcoeff_z, self.intercept_z = np.polyfit(2 * self.potim * self.nblock * 
                                                        np.arange(self.end - self.start), 
                                                        1E-5 * self.msd_z[self.start:self.end],
                                                        deg=1)
        return self.diffcoeff_z, self.intercept_z
    
    
    def plot_ensemble_msd_z(self):
        temp = self.prefix.split('/')[-2].split('.')[-1]
        plt.plot(self.timestep, self.msd_z[:], label=f"{temp}")

        if self.diffcoeff_z is not None:
            # visualize fitting line
            plt.plot(np.arange(self.start, self.end) * self.potim * self.nblock / 1000, 
                     self.diffcoeff_z*(2*self.potim*self.nblock*1.0E5)*
                     np.arange(self.end-self.start)+self.intercept_z*1.0E5, 
                     'k:')
            # visualize fitting range
            plt.axvline(self.start* self.potim * self.nblock / 1000, 0, 1, color='black', linewidth=0.5)
            plt.axvline(self.end* self.potim * self.nblock / 1000, 0, 1, color='black', linewidth=0.5)

        plt.title("z component", fontsize=14)
        plt.legend(loc='upper left')
        plt.xlabel('t (ps)', fontsize=13)
        plt.ylabel(r'MSD ($Å^2$)', fontsize=13)
    
    
    def save_msd(self):
        if not os.path.isdir('msd'):
            os.mkdir('msd')
        prefix = './msd/'
        timestep = self.timestep.reshape(self.timestep.shape[0], 1)
        msd = self.msd.reshape(self.msd.shape[0], 1)
        data_save = np.concatenate((timestep, msd), axis=1)

        temp = self.prefix.split('/')[-2].split('.')[-1]
        filename = prefix+'msd_'+temp+'.txt'
        np.savetxt(filename, data_save)



class getDiffusivity:
    def __init__(self, 
                 symbol,
                 path_xdatcar='./xdatcar',
                 skip=500,
                 segment=1,
                 start=500,
                 end='auto',
                 xyz=False,
                 label='auto',
                 temp='auto'):
        """
        Arg 1: (str) symbol; target atom (ex. 'O')
        Arg 2: (str) path_xdatcar; location of xdatcar folder
        Arg 3: (int; opt) skip; initial steps to be skipped (default=500)
        Arg 4: (int; opt) segment; number of segment (default=1)
        Arg 5: (int; opt) start; start step for fitting (default=500)
        Arg 6: (int) end; end; end step for fitting (default='auto')
        Arg 7: (bool; opt) xyz; if True, x, y, z component of D is calculated. (default=False)
        Arg 8: (list; opt) temp; range of temperature (default='auto')
        Arg 9: (list; opt) label; label of XDATCAR files (default='auto')
        """
        self.symbol = symbol
        self.prefix = path_xdatcar
        self.skip = skip
        self.segment = segment
        self.start = start
        
        # get temp
        if temp == 'auto':            
            self.temp = []
            for filename in os.listdir(self.prefix):
                if len(filename.split('.')) == 2:
                    first, second = filename.split('.')
                    if first == 'xdatcar':
                        t = second.split('K')[0]
                        self.temp += [t]
            self.temp = np.sort(np.array(self.temp, dtype=int))
        else:
            self.temp = np.array(temp, dtype=int)

        # get label
        self.label = {}
        if label == 'auto':
            for t in self.temp:
                label_t = []
                path_t = os.path.join(self.prefix, f"xdatcar.{t}K")
                for filename in os.listdir(path_t):
                    if len(filename.split('_')) == 2:
                        first, second = filename.split('_')
                        if first == 'XDATCAR':
                            label_t += [second]
                self.label[t] = label_t
        else:
            for t in self.temp:
                self.label[t] = label
        
        # automately set end
        if end == 'auto':
            path_example = os.path.join(self.prefix,
                                        f"xdatcar.{self.temp[0]}K",
                                        f"XDATCAR_{self.label[self.temp[0]][0]}",)
            with open(path_example, 'r') as file:
                lines = [line.strip() for line in file]
            for line in reversed(lines):
                contents = line.split()
                if contents[0] == 'Direct':
                    nsw = int(line.split('=')[-1])
                    break
            self.end = int((nsw - self.skip)/self.segment)
        else:
            self.end = end

        self.ensembles = []
        self.get_ensembles()

        self.diffcoeffs = []
        self.get_D()
        self.diffcoeffs = np.array(self.diffcoeffs)

        # get Ea and D0
        self.kb = 8.617332478E-5 # eV/K
        slop, intercept = np.polyfit(1/self.temp, np.log(self.diffcoeffs), deg=1)

        self.Ea = -slop * self.kb
        self.D0 = np.exp(intercept)

        self.plot_msd()
        self.plot_Arrhenius(self.diffcoeffs)
        self.save_D()

        print("")
        print("Ea = %.3f eV"%(self.Ea))
        print("D0 = %.3e m2/s"%(self.D0))

        if xyz:
            self.diffcoeffs_x = []
            self.diffcoeffs_y = []
            self.diffcoeffs_z = []
            self.get_Dx()
            self.get_Dy()
            self.get_Dz()
            self.plot_xyz()
            self.save_Dxyz()


    def get_ensembles(self):
        for t in tqdm(self.temp, 
                      bar_format='{l_bar}%s{bar:35}%s{r_bar}{bar:-10b}'% (Fore.GREEN, Fore.RESET),
                      ascii=False,
                      desc=f'{GREEN}{BOLD}TOTAL{RESET}'):
            ensemble = EnsembleEinstein(symbol=self.symbol,
                                        prefix=os.path.join(self.prefix,f"xdatcar.{t}K"),
                                        labels=self.label[t],
                                        segments=self.segment,
                                        skip=self.skip,
                                        start=self.start,
                                        end=self.end)
            self.ensembles += [ensemble]
            ensemble.save_msd()


    def get_D(self):
        self.diffcoeffs = [ensemble.diffcoeff for ensemble in self.ensembles]


    def get_Dx(self):
        self.diffcoeffs_x = [ensemble.diffcoeff_x for ensemble in self.ensembles]


    def get_Dy(self):
        self.diffcoeffs_y = [ensemble.diffcoeff_y for ensemble in self.ensembles]


    def get_Dz(self):
        self.diffcoeffs_z = [ensemble.diffcoeff_z for ensemble in self.ensembles]


    def plot_msd(self):
        for ensemble in self.ensembles:
            ensemble.plot_ensemble_msd()
            plt.legend(loc='upper left')
            plt.xlabel('t (ps)', fontsize=13)
            plt.ylabel(r'MSD ($Å^2$)', fontsize=13)
        plt.savefig('msd.png', dpi=300, transparent=False)
        plt.show()
        plt.close()


    def plot_Arrhenius(self, 
                       diffcoeffs, 
                       disp=True):
        kb = 8.617332478E-5 # eV/K
        slop, intercept = np.polyfit(1/self.temp, np.log(diffcoeffs), deg=1)
        tick = [r'$\frac{{1}}{{{}}}$'.format(t) for t in self.temp]
        Ea = -slop * kb
        D0 = np.exp(intercept)

        plt.plot(1/self.temp, np.log(diffcoeffs), 'bo')
        plt.plot(1/self.temp, slop*(1/self.temp)+intercept, 'k:')
        
        plt.xlabel('1/T (1/K)', fontsize=13)
        plt.xticks(1/self.temp, tick)
        plt.ylabel('ln D', fontsize=13)

        mid = int(len(self.temp)/2)
        plt.text(1/self.temp[mid]*1.05, 
                 slop*(1/self.temp[mid])+intercept, 
                 r"$E_a$=%.2f eV"%(Ea), 
                 fontsize=11)
        if disp:
            plt.savefig('Arrhenius.png', dpi=300, transparent=False)
            plt.show()
            plt.close()


    def save_D(self):
        with open('D.txt', 'w') as f:
            f.write(f"Ea = {self.Ea} eV\n")
            f.write(f"D0 = {self.D0} m2/s\n\n")
            f.write("T (K) \tD (m2/s)\n")
            for t, D in zip(self.temp, self.diffcoeffs):
                f.write(f"{t}\t{D}\n")


    def plot_xyz(self):
        figsize = (10,15)
        wspace = 0.4
        hspace = 0.5
        plt.figure(figsize=figsize)
        plt.subplots_adjust(wspace=wspace, hspace=hspace)
        
        # x
        plt.subplot(321)
        for ensemble in self.ensembles:
            ensemble.plot_ensemble_msd_x()
        
        plt.subplot(322)
        self.plot_Arrhenius(self.diffcoeffs_x, disp=False)

        # y
        plt.subplot(323)
        for ensemble in self.ensembles:
            ensemble.plot_ensemble_msd_y()
        
        plt.subplot(324)
        self.plot_Arrhenius(self.diffcoeffs_y, disp=False)

        # z
        plt.subplot(325)
        for ensemble in self.ensembles:
            ensemble.plot_ensemble_msd_z()
        
        plt.subplot(326)
        self.plot_Arrhenius(self.diffcoeffs_z, disp=False)

        plt.savefig('Dxyz.png', dpi=300, transparent=False)
        plt.show()
        plt.close()


    def save_Dxyz(self):
        with open('Dxyz.txt', 'w') as f:
            f.write("x-component\n")
            f.write("T (K) \tD (m2/s)\n")
            for t, D in zip(self.temp, self.diffcoeffs_x):
                f.write(f"{t}\t{D}\n")
            f.write("\n")

            f.write("y-component\n")
            f.write("T (K) \tD (m2/s)\n")
            for t, D in zip(self.temp, self.diffcoeffs_y):
                f.write(f"{t}\t{D}\n")
            f.write("\n")

            f.write("z-component\n")
            f.write("T (K) \tD (m2/s)\n")
            for t, D in zip(self.temp, self.diffcoeffs_z):
                f.write(f"{t}\t{D}\n")
            f.write("\n")



class CompareNEB:
    def __init__(self):
        self.kb = 8.617332478e-5 # eV/K

        # data from Einstein relation
        self.D_ER = []

        # site information
        self.site = []
        self.site_name = []
        self.site_E = []


    def add_Einstein(self, 
                     temp, 
                     D):
        if len(temp) != len(D):
            print("length of temp and D should be the same.")
            sys.exit(0)
        
        for t, d in zip(temp, D):
            dic_D = {}
            dic_D['T'] = t
            dic_D['D'] = d

            self.D_ER.append(dic_D)


    def print_Einstein(self):
        D_ER_sorted = sorted(self.D_ER,
                             key=lambda x:list(x.values()))
        
        print("T(K)\tD(m2/s)")
        for dic_D in D_ER_sorted:
            print("%d"%dic_D['T'], end='\t')
            print("%.2e"%dic_D['D'])


    def add_site(self, 
                 site, 
                 E):
        if len(site) != len(E):
            print("length of site and E should be the same.")
            sys.exit(0)
        
        for name, energy in zip(site, E):
            if name in self.site_name:
                print(f"{name} already exsits.")
                sys.exit(0)
            dic_site = {}
            dic_site['name'] = name
            dic_site['E'] = energy
            dic_site['d'] = []
            dic_site['Ea'] = []
            dic_site['z'] = []

            self.site_name.append(name)
            self.site_E.append(energy)
            self.site.append(dic_site)


    def add_path(self, 
                 init_site, 
                 distance, 
                 Ea, 
                 z):
        num_path = len(init_site)

        for i in range(num_path):
            if not init_site[i] in self.site_name:
                print(f"{init_site[i]} is not defined.")
                sys.exit(0)
            else:
                idx = self.site_name.index(init_site[i])
            
            self.site[idx]['d'].append(distance[i])
            self.site[idx]['Ea'].append(Ea[i])
            self.site[idx]['z'].append(z[i])
    
    
    def print_path(self):
        print("site\td(Å)\tEa(eV)\tE(eV)\tz")
        for dic_site in self.site:
            num_path = len(dic_site['d'])

            for i in range(num_path):
                print(f"{dic_site['name']}", end='\t')
                print("%.3f"%dic_site['d'][i], end='\t')
                print("%.2f"%dic_site['Ea'][i], end='\t')
                print("%.2f"%dic_site['E'], end='\t')
                print("%d"%dic_site['z'][i], end='\n')


    def get_D_neb(self, 
                  T, 
                  f=1e13):
        T = np.array(T, dtype=float)
        E = np.array(self.site_E, dtype=float).reshape(-1, 1)
        
        prob = np.exp(-E/(self.kb*T))
        partition = np.sum(prob, axis=0)
        prob /= partition

        D_site = np.zeros((len(self.site), len(T)))
        for i, dic_site in enumerate(self.site):
            num_path = len(dic_site['d'])
            D_i = 0
            for j in range(num_path):
                a = dic_site['d'][j] * 1e-10
                z = dic_site['z'][j]
                D_i += (1/6)*z*a**2*f * np.exp(-dic_site['Ea'][j]/(self.kb*T))
            D_site[i] = D_i

        return np.sum(prob * D_site, axis=0)
    
    
    def plot_neb(self, 
                 T, 
                 f=1e13, 
                 label=None):
        T = np.array(T)
        D_NEB = self.get_D_neb(T, f)
        plt.plot(1/T, np.log(D_NEB), 'k:', label=label)


    def plot_Einstein(self):
        # plot D from ER
        T_ER = np.array([dic['T'] for dic in self.D_ER])
        D_ER = np.array([dic['D'] for dic in self.D_ER])
        plt.scatter(1/T_ER, np.log(D_ER), 
                    facecolor='None', edgecolor='r', s=50, linewidth=1.5, 
                    label='Einstein')
        
        # x ticks for ER
        if len(T_ER) >= 3:
            ticks_idx = np.array([0, len(T_ER)/2, len(T_ER)-1], dtype=int)
            T_ticks = T_ER[ticks_idx]
            ticks = [f"1/{temp}" for temp in T_ticks]
            plt.xticks(1/T_ticks, ticks)
        else:
            ticks = [f"1/{temp}" for temp in T_ER]
            plt.xticks(1/T_ER, ticks)
        
        
    def plot_Arrhenius(self,
                       T,
                       f=[],
                       save=True,
                       outfig='Arrhenius.png',
                       dpi=300,
                       fmin=1e12,
                       fmax=1e14,
                       step=10000):
        # plot NEB
        label_NEB = None
        if len(f) == 0:
            f = [self.get_frequency(fmin, fmax, step)]
            label_NEB = "NEB " + "(" + "%.2e"%f[0] + " Hz)"

        if label_NEB is None:
            label_NEB = "NEB"

        for freq in f:
            self.plot_neb(T, freq, label=label_NEB)
            label_NEB = None
        
        # plot ER
        self.plot_Einstein()

        plt.xlabel('1/T (1/K)', fontsize=12)
        plt.ylabel('ln D', fontsize=12)
        plt.legend(fontsize=10.5)

        if save:
            plt.savefig(outfig, dpi=dpi)
        plt.show()


    def get_frequency(self,
                      fmin=1e12,
                      fmax=1e14,
                      step=10000):
        
        freq = np.linspace(fmin, fmax, step)
        T_ER = np.array([dic['T'] for dic in self.D_ER])
        D_ER = np.array([dic['D'] for dic in self.D_ER])

        err_min = np.inf
        f_opt = None
        for f in freq:
            D_NEB = self.get_D_neb(T_ER, f)
            err = np.sum((D_NEB - D_ER)**2)
            if err < err_min:
                err_min = err
                f_opt = f
                
        return f_opt



import os
import sys
import json
import math
import copy   
import numpy as np
import matplotlib.pyplot as plt

from PIL import Image
from tqdm import tqdm
from colorama import Fore
from tabulate import tabulate

from collections import defaultdict
from itertools import permutations

from pymatgen.core import Structure
from pymatgen.analysis.local_env import VoronoiNN
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

# For Arrow3D
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d

# color map for tqdm
BOLD = '\033[1m'
CYAN = '\033[36m'
MAGENTA = '\033[35m'
GREEN = '\033[92m' # Green color
RED = '\033[91m'   # Red color
RESET = '\033[0m'  # Reset to default color

class Arrow3D(FancyArrowPatch):
    def __init__(self, 
                 xs, 
                 ys, 
                 zs, 
                 *args, 
                 **kwargs):
        """
        helper class to drqw 3D arrows
        """
        super().__init__((0,0), (0,0), *args, **kwargs)
        self._verts3d = xs, ys, zs
        
    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        return np.min(zs)


class Lattice:
    def __init__(self, 
                 poscar_lattice, 
                 symbol,
                 rmax=3.0,
                 tol=1e-3,
                 tolerance=1e-3,
                 verbose=False):
        # read arguments
        self.poscar_lattice = poscar_lattice
        self.symbol = symbol
        self.rmax = rmax
        self.tol = tol
        self.tolerance = tolerance
        self.verbose = verbose
        
        # check error
        if not os.path.isfile(self.poscar_lattice):
            sys.exit(f"{self.poscar_lattice} is not found.")
            
        with open(self.poscar_lattice, 'r', encoding='utf-8') as f:
            contents = f.read()
            self.structure = Structure.from_str(contents, fmt='poscar')
        
        if not any(site.specie.symbol==self.symbol for site in self.structure):
            sys.exit(f"{self.symbol} is not in {self.poscar_lattice}")
        
        # attributes    
        self.path = []
        self.path_name = []
        self.site_name = None
        self.lattice_point = None
        self.lattice_parameter = self.structure.lattice.matrix
        self.find_hopping_path()
        
        # unknown path
        self.path_unknown = []
        
        # summary path
        if self.verbose:
            self.summary()
        
    def find_hopping_path(self):
        # find inequivalent sites
        sga = SpacegroupAnalyzer(self.structure)
        sym_structure = sga.get_symmetrized_structure()
        non_eq_sites = sym_structure.equivalent_sites
        non_eq_sites = [
            site_group for site_group in non_eq_sites if site_group[0].specie.symbol==self.symbol
            ]
        index = []
        for sites in non_eq_sites:
            index_sites = []
            for site in sites:
                coords = site.coords
                for i, _site in enumerate(self.structure.sites):
                    if np.linalg.norm(coords - _site.coords) < self.tolerance:
                        index_sites.append(i)
            index.append(index_sites)
        # index = np.array(index, dtype=int)
        
        # save site names
        self.site_name = [f"site{i+1}" for i in range(len(index))]
        
        # save lattice points
        self.lattice_point = []
        for i in range(min(map(min,index)), max(map(max,index))+1):
            for j, index_j in enumerate(index):
                if i in index_j:
                    site_i = j+1
                    break
            point = {}
            point['site'] = f"site{site_i}"
            point['coord'] = self.structure[i].frac_coords
            point['coord_C'] = self.structure[i].coords
            self.lattice_point.append(point)
            
        # find hopping paths
        nn_finder = VoronoiNN(tol=self.tol)
        self.path, self.path_name = [], []
        for i, idx in enumerate([index_i[0] for index_i in index]):
            paths_idx = []
            distances = np.array([], dtype=float)
            site_init = f"site{i+1}"
            neighbors = nn_finder.get_nn_info(self.structure, idx)
            neighbors = [
                neighbor for neighbor in neighbors if neighbor['site'].specie.symbol==self.symbol
                ]
            for neighbor in neighbors:
                distance = self.structure[idx].distance(neighbor['site'])
                if distance < self.rmax:
                    for j, index_j in enumerate(index):
                        if neighbor['site_index'] in index_j:
                            site_final = j+1
                            break
                    site_final = f"site{site_final}"
                    path_index = np.where(abs(distances - distance) < self.tolerance)[0]
                    if len(path_index) == 0:
                        path = {}
                        path['site_init'] = site_init
                        path['site_final'] = site_final
                        path['distance'] = float(distance)
                        path['z'] = 1
                        path['coord_init'] = self.structure[idx].frac_coords
                        path['coord_final'] = neighbor['site'].frac_coords
                        paths_idx.append(path)
                        distances = np.append(distances, distance)
                        self.path_name.append(f"{chr(i+65)}{len(paths_idx)}")
                    else:
                        paths_idx[path_index[0]]['z'] += 1
            self.path += paths_idx
        self.path = sorted(self.path, key=lambda x: (x['site_init'], x['distance']))
        self.path_name = sorted(self.path_name)
        for path, name in zip(self.path, self.path_name):
            path['name'] = name
    
    def summary(self):
        print(f"Number of inequivalent sites for {self.symbol} : {len(self.site_name)}")
        print(f"Number of inequivalent paths for {self.symbol} : {len(self.path_name)} (Rmax = {self.rmax:.2f} Å)")
        print('')
        print('Path information')
        headers = ['name', 'init', 'final', 'a(Å)', 'z', 'coord_init', 'coord_final']
        data = [
            [
                path['name'], 
                path['site_init'], 
                path['site_final'], 
                f"{path['distance']:.5f}", 
                path['z'],
                f"[{path['coord_init'][0]:.5f} {path['coord_init'][1]:.5f} {path['coord_init'][2]:.5f}]", 
                f"[{path['coord_final'][0]:.5f} {path['coord_final'][1]:.5f} {path['coord_final'][2]:.5f}]"
            ] for path in self.path
        ]
        print(tabulate(data, headers=headers, tablefmt="simple"))


class Trajectory:
    def __init__(self,
                 interval: float,
                 num_vac: int,
                 lattice,
                 pos_file: str,
                 force_file: str,
                 cond_file: str,
                 verbose: bool = True):
        
        self._validate_file(pos_file)
        self._validate_file(force_file)
        self._validate_file(cond_file)
        
        self.interval = interval
        self.num_vac = num_vac
        self.vervose = verbose
        self.cmap = self._custrom_cmap()
        
        # md conditions
        self.nsw = None  # nsw <- int(nsw / nblock)
        self.temp = None
        self.potim = None  # potim <- potim * nblock
        self.nblock = None
        self.symbol = None
        self.nsw_cut = None
        self.num_step = None
        self.num_atom = None
        self.interval_nsw = None
        self.read_cond_file(cond_file)
        
        # lattice information
        self.lattice = lattice
        self.lattice_point = None
        self.lattice_point_C = None
        self.lattice_parameter = None
        self.num_lattice_point = None
        self._init_lattice_point(lattice)
        
        self.occupation = None
        self.trace_arrows = None
        self.atomic_trajectory(pos_file, force_file)
        
        # trace arrows
        self.trace_arrows = None
        self.get_trace_arrows()
        
        # vacancy trajectory
        self.hopping_sequence = {}
        self.vacancy_trajectory_index = {}
        self.vacancy_trajectory_coord_C = {}
        self.transient_vacancy = {}
        self.get_vacancy_trajectory()
    
    def nearest_lattice_points(self, coords):  
        # coords: (num_atom, 3)
        diff = self.lattice_point[None, :, :] - coords[:, None, :]
        diff = diff - np.floor(diff)
        diff[diff > 0.5] -= 1.0
        diff[diff < -0.5] += 1.0
        diff_cart = np.tensordot(diff, self.lattice_parameter, axes=([2], [1]))
        norm = np.linalg.norm(diff_cart, axis=2)
        return np.argmin(norm, axis=1).astype(np.int16) 

    def atomic_trajectory(self, 
                          pos_file, 
                          force_file):
        pos = np.load(pos_file, mmap_mode='r')
        force = np.load(force_file, mmap_mode='r')
        
        check_init = False
        occupation = np.zeros((self.num_step, self.num_atom), dtype=np.int16)
        for i in range(self.num_step):
            # read step-wise data
            start = i * self.interval_nsw
            end = start + self.interval_nsw
            pos_chunk = np.average(pos[start:end], axis=0) # (num_atom, 3)
            force_chunk = np.average(force[start:end], axis=0) # (num_atom, 3)
            
            # proximity-based occupation
            occupation_i = self.nearest_lattice_points(pos_chunk)
            
            # initial occupation
            if not check_init:
                if len(set(occupation_i)) == self.num_atom:
                    for j in range(i+1):
                        occupation[j] = occupation_i
                    check_init = True
                continue
            
            # TS criterion
            indices_move_atom = np.where(occupation_i != occupation[i-1])[0]
            
            for index in indices_move_atom:
                site_init = occupation[i-1][index]
                site_final = occupation_i[index]
                force_atom = force_chunk[index]
                
                p_init = self.lattice_point[site_init]
                p_final = self.lattice_point[site_final]
                p_atom = pos_chunk[index]
                
                r_init = self.displacement_PBC(p_atom, p_init)
                r_final = self.displacement_PBC(p_atom, p_final)
                
                eps = 1e-12
                norm_f = np.linalg.norm(force_atom)
                norm_init = np.linalg.norm(r_init)
                norm_final = np.linalg.norm(r_final)
                
                if norm_f < eps or norm_init < eps:
                    cos_init = np.nan
                else:
                    cos_init = np.dot(force_atom, r_init) / (norm_f * norm_init)

                if norm_f < eps or norm_final < eps:
                    cos_final = np.nan
                else:
                    cos_final = np.dot(force_atom, r_final) / (norm_f * norm_final)
                    
                if np.isnan(cos_init) or np.isnan(cos_final):
                    print(f"WARNING: NaN in cos_init/final at step {i}")
                    
                if cos_init > cos_final:
                    occupation_i[index] = site_init
            occupation[i] = occupation_i
        self.occupation = occupation.T

    def get_trace_arrows(self):
        change_in_occ = np.diff(self.occupation, axis=1)
        move_atom, move_step = np.where(change_in_occ != 0)
        move_step += 1
        
        self.trace_arrows = {}
        for step, atom in zip(move_step, move_atom):
            arrow = {}
            arrow['c'] = self.cmap[(atom)%len(self.cmap)]
            arrow['lattice_point'] = [
                self.occupation[atom][step-1],
                self.occupation[atom][step]
            ]
            arrow['p'] = np.vstack((
                self.lattice_point_C[self.occupation[atom][step-1]],
                self.lattice_point_C[self.occupation[atom][step]]
            ))
            
            if step in self.trace_arrows.keys():
                self.trace_arrows[step].append(arrow)
            else:
                self.trace_arrows[step] = [arrow]
        
        for step in range(self.num_step):
            if not step in self.trace_arrows.keys():
                self.trace_arrows[step] = []              

    def get_vacancy_trajectory(self): 
        # initiallization
        self.hopping_sequence = {}
        self.vacancy_trajectory_index = {}
        self.vacancy_trajectory_coord_C = {}
        self.transient_vacancy = {0: np.array([], dtype=np.int16)}
        step_transient = {0: False}
        lattice_site = np.arange(self.num_lattice_point)
        
        # helper method to find vacancy paths
        def trace_paths(site_init, site_final, paths):
            path_map = defaultdict(list)
            for to_site, from_site in paths:
                path_map[from_site].append(to_site)

            site_final_set = set(site_final)
            candidate_routes = {s: [] for s in site_init}

            # Step 1: Collect all candidate paths for each site_init
            for s_init in site_init:
                stack = [(s_init, [s_init])]
                while stack:
                    current, route = stack.pop()
                    if current in site_final_set and not (len(route) == 1 and current == s_init):
                        candidate_routes[s_init].append(route)
                    for next_site in path_map.get(current, []):
                        if next_site not in route:
                            stack.append((next_site, route + [next_site]))

            # Step 2: Try all permutations of site_init to avoid final-site duplication
            for ordering in permutations(site_init):
                used_finals = set()
                results = []
                used_paths = set()
                for s in ordering:
                    found = False
                    for route in candidate_routes[s]:
                        if route[-1] not in used_finals:
                            results.append(route)
                            used_finals.add(route[-1])
                            # Track used path segments
                            for i in range(len(route) - 1):
                                # match input path format [to, from]
                                used_paths.add((route[i+1], route[i]))
                            found = True
                            break
                    if not found:
                        results.append(None)

                if len(results) == len(site_init) and None not in results:
                    reordered = [None] * len(site_init)
                    for i, s in enumerate(ordering):
                        idx = site_init.index(s)
                        reordered[idx] = results[i]
                    # Identify unused paths
                    path_set = set(map(tuple, paths))
                    unused_paths = list(path_set - used_paths)
                    return reordered, unused_paths

            # Fallback
            fallback = []
            used_paths = set()
            used_finals = set()
            for s in site_init:
                found = False
                for route in candidate_routes[s]:
                    if route[-1] not in used_finals:
                        fallback.append(route)
                        used_finals.add(route[-1])
                        for i in range(len(route) - 1):
                            used_paths.add((route[i+1], route[i]))
                        found = True
                        break
                if not found:
                    fallback.append(None)

            path_set = set(map(tuple, paths))
            unused_paths = list(path_set - used_paths)
            return fallback, unused_paths

        # vacancy trajectory at very first steps
        step_init = 0
        while True:
            site_vac = np.setdiff1d(lattice_site, self.occupation[:, step_init])
            if len(site_vac) == self.num_vac:
                break
            step_init += 1
        
        for step in range(step_init + 1):
            self.vacancy_trajectory_index[step] = copy.deepcopy(site_vac)
            self.vacancy_trajectory_coord_C[step] = self.lattice_point_C[site_vac]
        
        # find vacancy trajectory
        for step in range(step_init+1, self.num_step):
            site_vac_new = np.setdiff1d(lattice_site, self.occupation[:, step])
            
            # check presence of transient vacancy
            step_transient[step] = True if len(site_vac_new) > self.num_vac else False
            
            # check hopping events
            site_init = np.setdiff1d(site_vac, site_vac_new)
            site_final = np.setdiff1d(site_vac_new, site_vac)
            
            # no hop
            if len(site_init) == 0:
                self.vacancy_trajectory_index[step] = copy.deepcopy(site_vac)
                self.vacancy_trajectory_coord_C[step] = self.lattice_point_C[site_vac]
            
            # hops occur
            else:
                # check transient vacancies
                loop = 1
                site_transient = []
                paths= [arrow['lattice_point'] for arrow in self.trace_arrows[step]]
                while step_transient[step - loop]:
                    paths += [arrow['lattice_point'] for arrow in self.trace_arrows[step-loop]]
                    site_transient += list(self.transient_vacancy[step-loop])
                    loop += 1
                
                # get hoppigg paths
                site_final = np.array(list(set(list(site_final) + site_transient)))
                path_connect, unused_path = trace_paths(list(site_init), site_final, paths)
                
                for i, site in enumerate(site_init):
                    # find path index
                    path_index = None
                    for j, p in enumerate(path_connect):
                        if p is not None and p[0] == site:
                            path_index = j
                            break
                    
                    if path_index is None:
                        print(f"ERROR: Fail to find vacancy trajectory (step = {step}).")
                        print(f"site_init : {site_init}")
                        print(f"site_final : {site_final}")
                        print(f"paths : {paths}")
                        print(f"path_connect : {path_connect}")
                        sys.exit(0)
                    
                    # update vacancy site
                    site_vac[list(site_vac).index(site)] = path_connect[path_index][-1]
                    site_final = site_final[site_final != path_connect[path_index][-1]]
                        
                # compare 'updated site_vac' and 'site_vac_new' (correction for transient vac)
                site_remain = np.setdiff1d(site_vac, site_vac_new)
                if len(site_remain) > 0:
                    site_unexpect = np.setdiff1d(site_vac_new, site_vac)
                    path_unexpect, _ = trace_paths(list(site_remain), site_unexpect, unused_path)
                    
                    for i, site in enumerate(site_remain):
                        site_vac[list(site_vac).index(site)] = path_unexpect[i][-1]
                        site_final = site_final[site_final != path_unexpect[i][-1]]
                        
                        for path in path_connect:
                            if path is not None and path[-1] == site:
                                path.append(path_unexpect[i][-1])
                                break
                        
                # update vacancy trajectory
                self.hopping_sequence[step] = copy.deepcopy(path_connect)
                self.vacancy_trajectory_index[step] = copy.deepcopy(site_vac)
                self.vacancy_trajectory_coord_C[step] = self.lattice_point_C[site_vac]
                
            # update trasient vacancy
            self.transient_vacancy[step] = site_final
                   
    def _custrom_cmap(self):
        cmap = [
            'blue',
            'red',
            'teal',
            'indigo',
            'lime',
            'darkgoldenrod',
            'cyan',
            'hotpink',
            'dodgerblue',
            'dimgray',
            'forestgreen',
            'slateblue'
        ]
        return cmap  

    def _validate_file(self, file):
        if not os.path.isfile(file):
            print(f"ERROR: '{file}' is not found.")
            sys.exit(0)

    def _init_lattice_point(self, lattice):
        if lattice.symbol != self.symbol:
            print(f"Error: unmatched atomic symbol")
            sys.exit(0)
        
        self.lattice_point = np.array(
            [p['coord'] for p in lattice.lattice_point], dtype=np.float64
        )
        self.lattice_point_C = np.array(
            [p['coord_C'] for p in lattice.lattice_point], dtype=np.float64
        )
        self.num_lattice_point = len(self.lattice_point)
        self.lattice_parameter = lattice.lattice_parameter
        
    def read_cond_file(self, cond_file):
        with open(cond_file, "r") as f:
            condition = json.load(f)
        
        self.nsw = condition["nsw"]
        self.potim = condition["potim"]
        self.nblock = condition["nblock"]
        self.symbol = condition["symbol"]
        self.temp = condition["temperature"]
        self.num_atom = condition["atom_counts"][self.symbol]
        
        # effective nsw and potim
        self.nsw = int(self.nsw / self.nblock)
        self.potim *= self.nblock
        
        eps = 1e-9
        val = self.interval * 1000 / self.potim
        if math.isclose(val, round(val), abs_tol=eps):
             self.interval_nsw = int(round(val))
        else:
            print("ERROR: interval must be a multiple of 'potim * nblock'.")
            sys.exit(0)
            
        self.num_step = int(self.nsw / self.interval_nsw)
        self.nsw_cut = self.num_step * self.interval_nsw
    
    def distance_PBC(self, 
                     coord1: list, 
                     coord2: list):
        """
        Argument
        --------
        coord1 : list
            Initial coordinate in fraction (one point or multiple points)
        coord2 : list
            Final coordinate in fraction    
        """
        distance = coord1 - coord2
        distance[distance>0.5] -= 1.0
        distance[distance<-0.5] += 1.0

        if coord1.ndim == 1:
            return np.sqrt(np.sum(np.dot(distance, self.lattice_parameter)**2))
        else:
            return np.sqrt(np.sum(np.dot(distance, self.lattice_parameter)**2,axis=1))
    
    def displacement_PBC(self, 
                         r1, 
                         r2):
        disp = r2 - r1
        disp[disp > 0.5] -= 1.0
        disp[disp < -0.5] += 1.0
        return np.dot(disp, self.lattice_parameter)
    
    def plot_lattice(self, ax, label=False):
        coord_origin = np.zeros((1, 3))
        
        def plot_edge(start, end):
            edge = np.concatenate((start.reshape(1, 3), end.reshape(1, 3)), axis=0).T
            ax.plot(edge[0], edge[1], edge[2], 'k-', marker='none')

        a, b, c = self.lattice_parameter
        edges = [
            (coord_origin, a), (coord_origin, b), (coord_origin, c),
            (a + b, a), (a + b, b),
            (b + c, b), (b + c, c),
            (c + a, c), (c + a, a),
            (a + b + c, a + b), 
            (a + b + c, b + c),
            (a + b + c, c + a)
        ]

        for start, end in edges:
            plot_edge(start, end)

        ax.scatter(*self.lattice_point_C.T, facecolor='none', edgecolors='k', alpha=0.8)
        
        if label:
            for i, coord in enumerate(self.lattice_point_C):
                ax.text(*coord.T, s=f"{i}", fontsize='xx-small')

        ax.set_xlabel('x (Å)')
        ax.set_ylabel('y (Å)')
        ax.set_zlabel('z (Å)')

    def animation(self,
                  index: list = 'all',
                  step: list = 'all',
                  vac: bool = True,
                  gif: bool = True,
                  filename: str = 'traj.gif',
                  foldername: str = 'snapshot',
                  update_alpha: float = 0.75,
                  fps: int = 5,
                  loop = 0,
                  dpi: int = 100,
                  legend: bool = False,
                  label: bool = False):
        """
        Arguments
        ---------
        index : list or 'all', optional
            Index of atoms interested in (not index of lattice sites)
        step : list or 'all', optional
            Steps interested in.
        vac : bool, optional
            If true, vacancy is displayed.
        gif : bool, optional
            If true, gif file is generated.
        filename : list, optional
            Name of gif output file.
        foldername : list, optional
            Path of directory in which the snapshots save.
        update_alpha : float, optional
            Rate of increasing transparency of trace arrows.
        """
        if not os.path.isdir(foldername):
            os.mkdir(foldername)

        if index == 'all':
            index = np.arange(self.num_atom)
        
        if str(step) == 'all':
            step = np.arange(self.num_step)
        
        files = []
        for step in tqdm(step,
                         bar_format='{l_bar}%s{bar:35}%s{r_bar}{bar:-10b}'%(Fore.GREEN, Fore.RESET),
                         ascii=False,
                         desc=f'{RED}{BOLD}Progress{RESET}'):
            
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

            # plot lattice and lattice points
            self.plot_lattice(ax, label=label)

            # plot atoms
            for i, idx in enumerate(index):
                ax.scatter(*self.lattice_point_C[self.occupation][idx][step].T,
                           facecolor=self.cmap[idx%len(self.cmap)],
                           edgecolor='none',
                           alpha=0.8,
                           label=f"{idx}")
            
            # plot trace arrows
            alpha = 1
            for i in reversed(range(step+1)):
                for arrow in self.trace_arrows[i]:
                    arrow_prop_dict = dict(mutation_scale=10,
                                           arrowstyle='->',
                                           color=arrow['c'],
                                           alpha=alpha,
                                           shrinkA=0, 
                                           shrinkB=0)
                    disp_arrow = Arrow3D(*arrow['p'].T, **arrow_prop_dict)
                    ax.add_artist(disp_arrow)
                alpha *= update_alpha

            # plot vacancy
            if vac:
                # true vacancy
                ax.plot(*self.vacancy_trajectory_coord_C[step].T,
                        color='yellow', 
                        marker='o', 
                        linestyle='none', 
                        markersize=8, 
                        alpha=0.8, 
                        zorder=1)
                
                # trasient vacancy
                ax.plot(*self.lattice_point_C[self.transient_vacancy[step]].T,
                        color='orange', 
                        marker='o', 
                        linestyle='none', 
                        markersize=8, 
                        alpha=0.8, 
                        zorder=1)

            # make snapshot
            time = step * self.interval_nsw * self.potim / 1000 # ps
            time_tot = self.nsw_cut * self.potim / 1000 # ps
            plt.title("(%.2f/%.2f) ps, (%d/%d) step"%(time, time_tot, step, self.num_step))

            if legend:
                plt.legend()

            # save snapshot
            snapshot = os.path.join(foldername, f"snapshot_{step}.png")
            files.append(snapshot)
            plt.savefig(snapshot, dpi=dpi)
            plt.close()
        
        # make gif file
        if gif:
            print(f"Merging snapshots...")
            imgs = [Image.open(file) for file in files]
            imgs[0].save(fp=filename, format='GIF', append_images=imgs[1:], 
                         save_all=True, duration=int(1000/fps), loop=loop)
            print(f"{filename} was created.")


class TrajectoryAnalyzer:
    def __init__(self,
                 lattice,
                 trajectory,
                 tolerance=1e-3,
                 verbose=True):
        """
        Arguments
        ---------
        lattice : Lattice
            Lattice object
        trajectory : Trajectory
            Trajectory object
        tolerance : float, optional
            Tolerance for numerical accuracy
        verbose : bool, optoinal
            Verbosity flag
        """
        
        # argument
        self.trajectory = trajectory
        self.num_vac = trajectory.num_vac
        self.lattice = lattice
        self.tolerance = tolerance
        self.verbose = verbose
        
        # lattice information 
        self.path = lattice.path
        self.path_name = lattice.path_name
        self.site_name = lattice.site_name
        self.path_distance = np.array([p['distance'] for p in lattice.path])
        self.lattice_point = lattice.lattice_point
        
        # unknown path
        self.path_unknown = lattice.path_unknown
        
        # hopping history
        self.hopping_history = [[] for _ in range(self.num_vac)] # self.path_vac
        self.counts = np.zeros((self.num_vac, len(self.path_name)))
        self.hopping_statistics()
        
        # counts for unknown paths
        self.unknown_name = [f"unknown{i+1}" for i in range(len(self.path_unknown))]
        self.counts_unknown = np.zeros((self.num_vac, len(self.path_unknown)))
        self.counts_unknown_path()
        
        # random walk msd
        self.msd_rand = None # unknown path들도 포함됨
        self.random_walk_msd()
        
        # vacancy residence time (ps)
        self.residence_time = np.zeros((self.num_vac, len(self.site_name)))
        self.get_residence_time()
        
        # summary
        if verbose:
            self.summary()
        
    def hopping_statistics(self):
        for step, sequence in self.trajectory.hopping_sequence.items():
            for path in sequence:
                # check vacancy index
                index_vac = list(self.trajectory.vacancy_trajectory_index[step]).index(path[-1])
                
                for i in range(len(path) - 1):
                    index_init = path[i]
                    index_final = path[i+1]
                    
                    distance = self.trajectory.distance_PBC(
                        self.trajectory.lattice_point[index_init],
                        self.trajectory.lattice_point[index_final]
                    )
                    
                    # categorize migration
                    check_normal = False
                    for i, p in enumerate(self.path):
                        check1 = True if np.abs(p['distance'] - distance) < self.tolerance else False
                        check2 = True if p['site_init'] == self.lattice_point[index_init]['site'] else False
                        check3 = True if p['site_final'] == self.lattice_point[index_final]['site'] else False
                        
                        if check1 and check2 and check3:
                            check_normal = True
                            index_path = i
                            break
                    
                    # normal path
                    if check_normal:
                        path_info = copy.deepcopy(self.path[index_path])
                        path_info['step'] = step
                        path_info['index_init'] = index_init
                        path_info['index_final'] = index_final
                        self.hopping_history[index_vac].append(path_info)
                        self.counts[index_vac][index_path] += 1
                        
                    # unknown path
                    else:
                        check_unknown = False
                        for i, p in enumerate(self.path_unknown):
                            check1 = True if np.abs(p['distance'] - distance) < self.tolerance else False
                            check2 = True if p['site_init'] == self.lattice_point[index_init]['site'] else False
                            check3 = True if p['site_final'] == self.lattice_point[index_final]['site'] else False
                            
                            if check1 and check2 and check3:
                                check_unknown = True
                                index_path = i
                        
                        if check_unknown:
                            path_info = copy.deepcopy(self.path_unknown[index_path])
                            path_info['step'] = step
                            path_info['index_init'] = index_init
                            path_info['index_final'] = index_final
                            self.hopping_history[index_vac].append(path_info)
                        
                        else:
                            unknown_new={
                                'site_init': self.lattice_point[index_init]['site'],
                                'site_final': self.lattice_point[index_final]['site'],
                                'distance': distance,
                                'coord_init': self.lattice_point[index_init]['coord'],
                                'coord_final': self.lattice_point[index_final]['coord'],
                                'name': f"unknown{len(self.path_unknown)+1}"
                            }
                            self.path_unknown.append(copy.deepcopy(unknown_new))
                            unknown_new['step'] = step
                            unknown_new['index_init'] = index_init
                            unknown_new['index_final'] = index_final
                            self.hopping_history[index_vac].append(unknown_new)
    
    def get_residence_time(self):
        for indices in self.trajectory.vacancy_trajectory_index.values():
            for i, index in enumerate(indices):
                index_site = self.site_name.index(self.lattice.lattice_point[index]['site'])
                self.residence_time[i][index_site] += 1
        self.residence_time *= self.trajectory.interval
        
    def counts_unknown_path(self):
        for index_vac in range(self.num_vac):
            for path in self.hopping_history[index_vac]:
                if 'unknown' in path['name']:
                    index_path = int(''.join(filter(str.isdigit, path['name']))) - 1
                    self.counts_unknown[index_vac][index_path] += 1
    
    def random_walk_msd(self):
        distance_all = np.array(
            list(self.path_distance) + [p['distance'] for p in self.path_unknown]
        )
        counts_all = np.hstack((self.counts, self.counts_unknown))
        self.msd_rand = np.average(
            np.sum(distance_all ** 2 * counts_all, axis=1)
        )
                
    def summary(self):
        # path counts
        name_all = self.path_name + self.unknown_name
        counts_all = np.hstack((self.counts, self.counts_unknown))
        counts_all = np.array(counts_all, dtype=np.int32)
        vacancy_name = [f"Vacancy{i+1}" for i in range(self.num_vac)]
        print("# Path counts :")
        header = ['path'] + vacancy_name
        data = np.vstack((name_all, counts_all)).T
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        # vacancy residence time
        print("# Vacancy residence time (ps) :")
        header = ['site'] + vacancy_name
        data = np.vstack((self.site_name, self.residence_time)).T
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        # hopping sequence
        print("# Hopping sequence :")
        for i in range(self.num_vac):
            print(f"# Vacancy{i+1}")
            header = ['num', 'time (ps)', 'path', 'a (Å)', 'initial site', 'final site']
            data = [
                [
                    f"{j+1}",
                    f"{path['step'] * self.trajectory.interval:.2f}",
                    f"{path['name']}",
                    f"{path['distance']:.5f}",
                    f"{path['site_init']} [{', '.join(f'{x:.5f}' for x in self.trajectory.lattice_point[path['index_init']])}]",
                    f"{path['site_final']} [{', '.join(f'{x:.5f}' for x in self.trajectory.lattice_point[path['index_final']])}]"
                ] for j, path in enumerate(self.hopping_history[i])
            ]
            print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
            print('')

        
class Encounter:
    def __init__(self,
                 analyzer,
                 use_incomplete_encounter: bool = False,
                 verbose: bool = True):
        """
        Arguments
        ---------
        analyzer : TrajectoryAnalyzer
            TrajectoryAnalyzer object
        use_incomplete_encounter : bool, optional
            If true, incomplete encounters are used together in computations.
        verbose : bool, optional
            Verbosity flag
        """
        # arguments
        self.analyzer = analyzer
        self.trajectory = analyzer.trajectory
        self.tolerance = analyzer.tolerance
        self.use_incomplete_encounter = use_incomplete_encounter
        self.verbose = verbose
        
        # path information
        self.path = analyzer.path + analyzer.path_unknown
        self.path_name = [p['name'] for p in self.path]
        
        # unwrapped vacancy trajectory
        self.vacancy_coord_unwrap = None
        self.unwrapped_vacancy_trajectory()
        
        # encounter
        self.encounter_complete = []
        self.encounter_in_process = []
        self.encounter_PBC()
        self.num_encounter_complete = len(self.encounter_complete)
        self.num_encounter_incomplete = len(self.encounter_in_process)
        
        # encounter information
        self.msd = None
        self.path_count = None
        self.f_cor = None
        self.path_distance = None
        
        if use_incomplete_encounter:
            self.encounter_complete += self.encounter_in_process
        self.num_encounter = len(self.encounter_complete)
        
        if self.num_encounter == 0:
            if verbose:
                print("All encounters are incompleted")
        else:
            self.analyze_encounter()
            self.correlation_factor()
            if verbose:
                self.summary()
               
    def find_path_name(self, 
                    index_init: int, 
                    index_final: int):
        """
        Argument
        --------
        index_init : int
            Index of initial site
        index_final : int
            Index of final site
        """
        lattice_point = self.trajectory.lattice.lattice_point
        
        site_init = lattice_point[index_init]['site']
        site_final = lattice_point[index_final]['site']
        
        coord_init = lattice_point[index_init]['coord']
        coord_final = lattice_point[index_final]['coord']
        distance = self.trajectory.distance_PBC(coord_init, coord_final)
        
        for path in self.path:
            check1 = True if path['site_init'] == site_init else False
            check2 = True if path['site_final'] == site_final else False
            check3 = True if abs(path['distance'] - distance) < self.tolerance else False
            
            if check1 and check2 and check3:
                path_name = path['name']
                break

        return path_name
        
    def unwrapped_vacancy_trajectory(self):
        coords = self.trajectory.lattice_point[
            np.array(
                list(self.trajectory.vacancy_trajectory_index.values())
            )
        ]
        displacement = np.zeros_like(coords)
        displacement[1:] = np.diff(coords, axis=0)
        displacement[displacement > 0.5] -= 1.0
        displacement[displacement < -0.5] += 1.0
        displacement = np.cumsum(displacement, axis=0)
        self.vacancy_coord_unwrap = coords[0] + displacement
        
    def encounter_PBC(self):
        for step, sequence in self.trajectory.hopping_sequence.items():
            for path_connect in sequence:
                # vacancy index
                index_vac = list(
                    self.trajectory.vacancy_trajectory_index[step]
                ).index(path_connect[-1])
                
                # decompose path
                trace_arrow = [
                    path_connect[i-1:i+1][::-1] for i in range(len(path_connect)-1, 0, -1)
                ]
                
                # encounter analysis
                coord_init = self.vacancy_coord_unwrap[step][index_vac]
                for path in trace_arrow:
                    # atom index
                    try:
                        index_atom = list(self.trajectory.occupation[:, step]).index(path[-1])
                        
                    # transient vacancy may cause VauleError
                    except ValueError:
                        loop = 1
                        while True:
                            if len(self.trajectory.transient_vacancy[step-loop]) == 0:
                                print(f"Atomic index is not defined (step : {step}).")
                                sys.exit(0)
                                
                            match = next(
                                (arrow for arrow in self.trajectory.trace_arrows[step-loop]
                                if path == arrow['lattice_point']),
                                None
                            )
                            
                            if match:
                                index_atom = list(
                                    self.trajectory.occupation[:, step-loop]
                                ).index(path[-1])
                                break
                            
                            loop += 1

                    # path name
                    path_name = self.find_path_name(*path)
                    
                    # unwrapped atomic coord
                    displacement = np.diff(self.trajectory.lattice_point[path], axis=0)
                    displacement[displacement > 0.5] -= 1.0
                    displacement[displacement < -0.5] += 1.0
                    coord_final = coord_init + displacement.flatten()
                    
                    # comparison with existing encounters
                    index_encounter = None
                    for i, enc in enumerate(self.encounter_in_process):
                        if enc['index_atom'] == index_atom:
                            index_encounter = i
                            break
                    
                    # case 1. no matching encounter
                    if index_encounter is None:
                        encounter = {
                            'index_atom': index_atom,
                            'index_vac': index_vac,
                            'coord_init': coord_init,
                            'coord_final': coord_final,
                            'hopping_history': [path_name]
                        }
                        self.encounter_in_process.append(encounter)
                        coord_init = coord_final
                        continue
                    
                    # matching encounter
                    encounter_match = self.encounter_in_process[index_encounter]
                    coord_encounter = encounter_match['coord_final']
                    
                    distance = np.linalg.norm(
                        np.dot(
                            coord_encounter - coord_init,
                            self.trajectory.lattice_parameter
                        )
                    )
                    
                    # case 2. exactly matching encounter
                    if distance < self.tolerance:
                        # exchange with the associated vacancy : update encoutner
                        if encounter_match['index_vac'] == index_vac:
                            encounter_match['coord_final'] = coord_final
                            encounter_match['hopping_history'].append(path_name)
                        
                        # exchange with a new vacancy : terminate encounter
                        else:
                            # terminate the existing encounter
                            self.encounter_complete.append(encounter_match.copy())
                            del self.encounter_in_process[index_encounter]
                            
                            # initiate a new encounter
                            encounter = {
                                'index_atom': index_atom,
                                'index_vac': index_vac,
                                'coord_init': coord_init,
                                'coord_final': coord_final,
                                'hopping_history': [path_name]
                            }
                            self.encounter_in_process.append(encounter)
                    
                    # case 3. PBC matching encounter:
                    else:
                        # terminate the existing encounter
                        self.encounter_complete.append(encounter_match.copy())
                        del self.encounter_in_process[index_encounter]
                        
                        # initiate a new encounter
                        encounter = {
                            'index_atom': index_atom,
                            'index_vac': index_vac,
                            'coord_init': coord_init,
                            'coord_final': coord_final,
                            'hopping_history': [path_name]
                        }
                        self.encounter_in_process.append(encounter)
                    
                    coord_init = coord_final
                               
    def analyze_encounter(self):
        displacement = []
        self.path_count = np.zeros_like(self.path_name, dtype=np.float64)
        
        for encounter in self.encounter_complete:
            # path count
            for name in encounter['hopping_history']:
                index_path = self.path_name.index(name)
                self.path_count[index_path] += 1
            
            # displacement
            disp = encounter['coord_final'] - encounter['coord_init']
            disp = np.dot(disp, self.trajectory.lattice_parameter)
            displacement.append(disp)
            
        displacement = np.array(displacement)
        self.msd = np.average(np.sum(displacement**2, axis=1))
            
    def correlation_factor(self):
        self.path_distance = np.array([path['distance'] for path in self.path])
        denominator = np.sum(self.path_distance ** 2 * (self.path_count / self.num_encounter))
        self.f_cor = self.msd / denominator

    def summary(self):
        print("# Encounter analysis")
        print(f"  Use_incomplete_encounter : {self.use_incomplete_encounter}")
        print(f"  Correlation factor : {self.f_cor:.5f}")
        print(f"  Mean squared displacement (MSD) : {self.msd:.5f} Å2")
        print(f"  Number of complete encounters : {self.num_encounter_complete}")
        print(f"  Number of incomplete encounters : {self.num_encounter_incomplete}")
        print(f"  Number of encounters used in computations: {self.num_encounter}")
        print(f"  Total counts of hopping events : {int(np.sum(self.path_count))}")
        print(f"  Mean counts of hopping evnets : {np.sum(self.path_count) / self.num_encounter:.5f}")
        print('') 
        print(f"# Pathwise counts in encounters : ")
        header = ['path', 'a (Å)', 'count', 'count/enc']
        data = [
            [
                name,
                f"{a:.5f}",
                f"{int(count)}",
                f"{count / self.num_encounter:.5f}"
            ]
            for name, a, count in zip(self.path_name, self.path_distance, self.path_count)
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')      
                    



import os
import sys
import argparse
from fractions import Fraction
from vachoppy.inout import DataInfo
from vachoppy.core import *
from vachoppy.utils import *

try:
    from mpi4py import MPI
    PARALELL = True
except:
    PARALELL = False
    
BOLD = '\033[1m'
CYAN = '\033[36m'
MAGENTA = '\033[35m'
GREEN = '\033[92m'
RED = '\033[91m'
RESET = '\033[0m'

parser = argparse.ArgumentParser(
    description='VacHopPy: A Python Code to Analyze Vacancy Hopping',
    formatter_class=argparse.RawTextHelpFormatter
    )

group = parser.add_mutually_exclusive_group(required=True)

# key functionalities
group.add_argument(
    '-m', '--mode', 
    choices=['t', 'p', 'pp', 'f', 'e'], 
    help=(
        """Choose mode:
        't'  - Make animation for vacancy trajectory
        'p'  - Determine vacancy trajectory and calculate effectie hopping parameters (excluding z and nu)
        'pp' - Calculate effective z and nu (postprocess of p option)
        'f'  - Fingerpring analysis : assess lattice stability or phase transition
        'e'  - Calculate diffusion coefficient using Einstein relation (under test)
        """
        )
    )

# utilities
group.add_argument(
    '-u',
    choices=['extract_input', 'combine_vasprun', 'crop_vasprun', 'fingerprint', 'cosine_distance'],
    dest='util',
    help=(
        """Choose mode:
        'extract_input'   - Extract input files from vasprun.xml
        'combine_vasprun' - Combine two successive vasprun.xml
        'crop_vasprun'.   - Crop vasprun.xml
        'fingerprint'     - Extract fingerprint
        'cosine_distance' - Calculate cosine distance
        """
        )
)

check_mode = True if '-m' in sys.argv or '--mode' in sys.argv else False
check_util = True if '-u' in sys.argv else False

if check_mode:
    if '-m' in sys.argv:
        mode_index = sys.argv.index('-m') + 1
    elif '--mode' in sys.argv:
        mode_index = sys.argv.index('--mode') + 1

if check_util:
    mode_index = sys.argv.index('-u') + 1
    if mode_index < len(sys.argv):
        mode_value = sys.argv[mode_index]
        
        if mode_value == 'extract_input':
            parser.add_argument('symbol',
                                type=str,
                                help='symbol of atom species')
            parser.add_argument('-v', '--vasprun',
                                default='vasprun.xml',
                                help='input vasprun.xml file (default: vasprun.xml)')
            
        if mode_value == 'combine_vasprun':
            parser.add_argument('-v1', '--vasprun_in1',
                                required=True,
                                help='first vasprun.xml file')
            parser.add_argument('-v2', '--vasprun_in2',
                                required=True,
                                help='second vasprun.xml file')
            parser.add_argument('-out', '--vasprun_out',
                                default='vasprun_combined.xml',
                                help='combined vasprun.xml file (default: vasprun_combined.xml)')
        
        if mode_value == 'crop_vasprun':
            parser.add_argument('-n', '--nsw',
                                type=int,
                                required=True,
                                help='number of AIMD iterations to retain from the start')
            parser.add_argument('-v', '--vasprun_in',
                                default='vasprun.xml',
                                help='vasprun.xml file (default: vasprun.xml)')
            parser.add_argument('-out', '--vasprun_out',
                                default='vasprun_cropped.xml',
                                help='cropped vasprun.xml file (default: vasprun_cropped.xml)')
            
            
        if mode_value == 'cosine_distance':
            parser.add_argument('-f1', '--fingerprint_in1',
                                required=True,
                                help='first fingerprint file')
            parser.add_argument('-f2', '--fingerprint_in2',
                                required=True,
                                help='second fingerprint file')
            
        if mode_value == 'fingerprint':
            parser.add_argument('poscar',
                                type=str,
                                help='POSCAR to be used for fingerprint extraction')
            parser.add_argument('Rmax',
                                type=float,
                                help='maxiumu distance range (Å)')
            parser.add_argument('delta',
                                type=float,
                                help='bin size (Å)')
            parser.add_argument('sigma',
                                type=float,
                                help='sigma for Gaussian-smeared delta function')
            parser.add_argument('--prefix',
                                type=str,
                                default='fingerprint',
                                help='prefix for output files (default: fingerprint)')
            parser.add_argument('-d', '--disp',
                                action='store_true',
                                help='if use, fingerprint plot will be displayed')
        

if check_mode:
    if mode_index < len(sys.argv):
        mode_value = sys.argv[mode_index]

        if mode_value in ['p', 't', 'e']:
            parser.add_argument('-p1', '--prefix1', 
                                default='traj', 
                                help='name of outer directory (default: traj)')
            parser.add_argument('-p2', '--prefix2', 
                                default='traj', 
                                help='prefix of inner directories, ex.{prefix2}.{temp}K (default: traj)')
            
        if mode_value == 't':
            parser.add_argument('interval',
                                type=float,
                                help='time interval for averaging in ps')
            parser.add_argument('temp',
                                type=int,
                                help='temperatue in K')
            parser.add_argument('label',
                                type=str,
                                help='labels')
            parser.add_argument('-l', '--lattice',
                                type=str,
                                default='POSCAR_LATTICE',
                                help='lattice file in POSCAR format (default: POSCAR_LATTICE)')
            parser.add_argument('--update_alpha',
                                type=float,
                                default=0.75,
                                help='adjust the remaining time of the arrow (default: 0.75)')
            parser.add_argument('--show_index',
                                action='store_true',
                                help='if use, index of each atom will be shown')
            parser.add_argument('--dpi',
                                type=int,
                                default=300,
                                help='adjust dpi of snapshots (default: 300)')
            parser.add_argument('-v', '--verbose',
                                action='store_true',
                                help='verbosity for parameter calculation')
            
        if mode_value == 'p':
            parser.add_argument('interval',
                                type=float,
                                help='time interval for averaging in ps')
            parser.add_argument('-l', '--lattice',
                                type=str,
                                default='POSCAR_LATTICE',
                                help='lattice file in POSCAR format (default: POSCAR_LATICE)')
            parser.add_argument('--parallel',
                                action='store_true',
                                help='do parallel calculations (default: False)')
            parser.add_argument('--rmax',
                                type=float,
                                default=3.0,
                                help='maximum distance for hopping path identification (default: 3.0)')
            parser.add_argument('--tol',
                                type=float,
                                default=1e-3,
                                help='tolerance for VoronoiNN (default: 1e-3)')
            parser.add_argument('--tolerance',
                                type=float,
                                default=1e-3,
                                help='tolerance for distance comparison (default: 1e-3)')
            parser.add_argument('--use_complete_encounter',
                                action='store_true',
                                help='')
            parser.add_argument('--inset_correlatoin_factor',
                                action='store_true',
                                help='')
            
        if mode_value == 'pp':
            parser.add_argument('-p', '--parameter',
                                type=str,
                                default='parameter.txt',
                                help='parameter.txt file (default: parameter.txt)')
            parser.add_argument('-n', '--neb',
                                type=str,
                                default='neb.csv',
                                help='neb.csv file containing hopping barriers for each path (default: neb.csv)')
            
        if mode_value == 'f':
            parser.add_argument('interval',
                                type=float,
                                help='time interval for averaging in ps')
            parser.add_argument('Rmax',
                                type=float,
                                help='maxiumu distance range (Å)')
            parser.add_argument('delta',
                                type=float,
                                help='bin size (Å)')
            parser.add_argument('sigma',
                                type=float,
                                help='sigma for Gaussian-smeared delta function')
            parser.add_argument('-v','--vasprun',
                                type=str,
                                default='vasprun.xml',
                                help='path to vaprun.xml file (default: vasprun.xml)')
            parser.add_argument('-p','--poscar_ref',
                                default='POSCAR_REF',
                                type=str,
                                help='path to POSCAR of reference phase (default: POSCAR_REF)')
            parser.add_argument('--parallel',
                                action='store_true',
                                help='do parallel calculations (default: False)')
            parser.add_argument('--prefix1',
                                type=str,
                                default='snapshots',
                                help='directory to save xdatcar snapshots (default: snapshots)')
            parser.add_argument('--prefix2',
                                type=str,
                                default='fingerprints',
                                help='directory to save fingerprints (default: fingerprints)')
        
        if mode_value == 'e':
            parser.add_argument('t_width',
                                type=float,
                                help='x-range in msd plot in ps')
            parser.add_argument('--skip',
                                type=float,
                                default=0,
                                help='steps to be skipped in ps (default: 0)')
            parser.add_argument('--x_vac',
                                type=str,
                                default=1,
                                help='fraction of vacancy (use 1 for atom) (default: 1)')
            parser.add_argument('--start',
                                type=float,
                                default=1,
                                help='initial time used for linear fit (default: 1)')


args = parser.parse_args()

def main():
    if check_mode:
        # print calculation conditions
        if PARALELL:
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
        else:
            rank = 0
        
        if rank==0:
            print(f'{CYAN}{BOLD}VacHopPy is in progress{RESET}')
            # save arguments
            with open('arg.txt', 'w') as f:
                print(f'{GREEN}{BOLD}Arguments and Values :{RESET}')
                f.write('Arguments and Values :\n')
                for arg, value in vars(args).items():
                    print(f'    {arg} = {value}')
                    f.write(f'    {arg} = {value}\n')
            print('')
        
        # functionalities
        if mode_value in ['p', 't', 'e']:
            data = DataInfo(
                prefix1=args.prefix1,
                prefix2=args.prefix2,
                verbose=True
            )
        
        if mode_value == 't':
            traj = MakeAnimation(
                data=data,
                temp=args.temp,
                label=args.label,
                interval=args.interval,
                poscar_lattice=args.lattice,
                update_alpha=args.update_alpha,
                show_index=args.show_index,
                dpi=args.dpi,
                verbose=args.verbose
            )
            
        if mode_value == 'p':
            effective_params = EffectiveHoppingParameter(
                data=data,
                interval=args.interval,
                poscar_lattice=args.lattice,
                parallel=args.parallel,
                file_out='parameter.txt',
                rmax=args.rmax,
                tol=args.tol,
                tolerance=args.tolerance,
                use_incomplete_encounter=not(args.use_complete_encounter),
                inset_correlatoin_factor=args.inset_correlatoin_factor,
                verbose=True
            )
        
        if mode_value == 'pp':
            post = PostEffectiveHoppingParameter(
                file_params=args.parameter,
                file_neb=args.neb
            )
            
        if mode_value == 'f':
            phase = PhaseTransition(
                vasprun=args.vasprun,
                interval=args.interval,
                Rmax=args.Rmax,
                delta=args.delta,
                sigma=args.sigma,
                parallel=args.parallel,
                poscar_ref=args.poscar_ref,
                prefix1=args.prefix1,
                prefix2=args.prefix2
            )
        
        if mode_value == 'e':
            msd = MSD(
                data=data, 
                tmax=args.t_width, 
                skip=args.skip,
                start=args.start,
                symbol=args.symbol, 
                x_vac=float(Fraction(args.x_vac))
            )
            
    if check_util:
        if mode_value == 'extract_input':
            extract_from_vasp(args.symbol, vasprun=args.vasprun)
            
        if mode_value == 'combine_vasprun':
            combine_vasprun(args.vasprun_in1, args.vasprun_in2, args.vasprun_out)
            print(f'{args.vasprun_out} is created')
            
        if mode_value == 'crop_vasprun':
            crop_vasprun(args.vasprun_in, args.nsw, args.vasprun_out)
            print(f'{args.vasprun_out} is created')
        
        if mode_value == 'cosine_distance':
            GetCosineDistance(args.fingerprint_in1, args.fingerprint_in2)
            
        if mode_value == 'fingerprint':
            finger = GetFingerPrint(
                poscar=args.poscar,
                Rmax=args.Rmax,
                delta=args.delta,
                sigma=args.sigma,
                prefix=args.prefix,
                disp=args.disp
            )


import os
import sys
import time
from tqdm import tqdm
from colorama import Fore
from tabulate import tabulate
from vachoppy.parameter import *

try:
    from mpi4py import MPI
    PARALELL = True
except:
    PARALELL = False

# color map for tqdm
BOLD = '\033[1m'
CYAN = '\033[36m'
MAGENTA = '\033[35m'
GREEN = '\033[92m' # Green color
RED = '\033[91m'   # Red color
RESET = '\033[0m'  # Reset to default color


class DataInfo:
    def __init__(self, 
                 prefix1: str ='traj',
                 prefix2: str ='traj',
                 prefix_pos: str = 'pos',
                 prefix_force: str = 'force',
                 prefix_cond: str = 'cond',
                 verbose: bool =True):
        """
        Args:
            prefix1 (str, optional): 
                name of outer directory. Defaults to 'traj'.
            prefix2 (str, optional): 
                prefix for subdirectories. Defaults to 'traj'.
            prefix_pos (str, optional): 
                prefix for pos.npy. Defaults to 'pos'.
            prefix_force (str, optional): 
                prefix for force.npy. Defaults to 'force'.
            prefix_cond (str, optional): 
                prefix for cond.json. Defaults to 'cond'.
            verbose (bool, optional): 
                verbosity flag. Defaults to True.
        """
        
        self.prefix1 = prefix1
        self.prefix2 = prefix2
        self.prefix_pos = prefix_pos
        self.prefix_cond = prefix_cond
        self.prefix_force = prefix_force
        
        # subdirectories
        self.sub_directory = []
        self.get_sub_directory()
        
        # labels
        self.label = []
        self.get_label()
        
        # integrity of input files
        self.temp = []
        self.pos = []
        self.cond = []
        self.force = []
        self.check_integrity()
        
        
        self.datainfo = [
            [temp, label] for i, temp in enumerate(self.temp)
            for label in self.label[i]
        ]
        
        self.label_all = list(
            {element for sublabel in self.label for element in sublabel}
        )
        self.label_all.sort()
        
        if verbose:
            with open('data.txt', 'w') as f:
                original_stdout = sys.stdout
                sys.stdout = f
                try:
                    self.summary()
                finally:
                    sys.stdout = original_stdout
        
    def get_sub_directory(self):
        list_dir = os.listdir(os.path.join(os.getcwd(), self.prefix1))
        self.sub_directory = [d for d in list_dir if self.prefix2 in d]
        self.sub_directory.sort()
        
    def get_label(self):
        for dir in self.sub_directory:
            # gather cond.json file
            label_i = [
                "".join(f.split(".")[:-1]).split("_")[-1] 
                for f in os.listdir(os.path.join(self.prefix1, dir))
                if f.split("_")[0] == self.prefix_cond
                if f.split(".")[-1] == "json"
            ]
            label_i.sort()
            self.label.append(label_i)
            
    def check_integrity(self):
        for i, dir in enumerate(self.sub_directory):
            temp_i = []
            pos_i = []
            cond_i = []
            force_i = []
            for label in self.label[i]:
                pos_file = os.path.join(
                    self.prefix1, dir, f"{self.prefix_pos}_{label}.npy"
                )
                cond_file = os.path.join(
                    self.prefix1, dir, f"{self.prefix_cond}_{label}.json"
                )
                force_file = os.path.join(
                    self.prefix1, dir, f"{self.prefix_force}_{label}.npy"
                )
                
                # check pos.npy, force.npy
                integrity = True
                if not os.path.isfile(pos_file):
                    print(f"Warning: file absent ({pos_file})")
                    integrity = False
                    
                if not os.path.isfile(force_file):
                    print(f"Warning: file absent ({force_file})")
                    integrity = False
                
                if integrity:
                    with open(cond_file, "r") as f:
                        cond = json.load(f)
                    temp_i.append(cond['temperature'])
                    pos_i.append(pos_file)
                    cond_i.append(cond_file)
                    force_i.append(force_file)
                else:
                    self.label[i].remove(label)
                    
            # check temperature
            temp_first = temp_i[0]
            check_temp = all(abs(x - temp_first) < 1e-6 for x in temp_i[1:])
            if check_temp:
                self.temp.append(temp_first)
                self.pos.append(pos_i)
                self.cond.append(cond_i)
                self.force.append(force_i)
            else:
                print(f"Error: temperatures within the same directory must be the same ({dir})")
            
    def summary(self):
        print("# Sumamry of input data:")
        header = ['label'] + [f"{temp}K" for temp in self.temp]
        data = [
            [label] + ['O' if label in self.label[j] else 'X' for j in range(len(self.temp))]
            for label in self.label_all
        ]
        data.append(
            ['Total'] + [len(label) for label in self.label]
        )
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        print("# List of cond.json")
        for i, temp in enumerate(self.temp):
            print(f"{temp} K :")
            for file in self.cond[i]:
                print(f"  {file}")
        print("")
        
        print("# List of pos.npy")
        for i, temp in enumerate(self.temp):
            print(f"{temp} K :")
            for file in self.pos[i]:
                print(f"  {file}")
        print("")
        
        print("# List of force.npy")
        for i, temp in enumerate(self.temp):
            print(f"{temp} K :")
            for file in self.force[i]:
                print(f"  {file}")
        print("")


class VacancyInfo:
    def __init__(self,
                 data,
                 poscar_lattice: str = "POSCAR_LATTICE"):
        """
        Args:
            data (DataInfo): 
                DataInfo object
            poscar_lattice (str): 
                path to POSCAR_LATTICE. Defaults to 'POSCAR_LATTICE'.
        """
        
        self.data = data
        self.poscar_lattice = poscar_lattice
        
        self.lattice_poscar = []
        self.atom_counts_poscar = {}
        self.read_poscar()
        
        self.symbol_vac = None
        self.number_vac = None
        self.read_cond()
        
    def read_poscar(self):
        check_symbol, check_number = False, False
        with open(self.poscar_lattice, 'r') as f:
            for i, line in enumerate(f):
                if i == 2:
                    self.lattice_poscar.append(line.split()[:3])
                    
                if i == 3:
                    self.lattice_poscar.append(line.split()[:3])
                    
                if i == 4:
                    self.lattice_poscar.append(line.split()[:3])
                
                if i == 5:
                    atom_symbol = line.split()
                    check_symbol = True
                    
                if i == 6:
                    atom_number = list(map(int, line.split()))
                    check_number = True
                    
                if check_symbol and check_number:
                    break
                
        self.lattice_poscar = np.array(self.lattice_poscar, dtype=np.float64)
        
        for symbol, number in zip(atom_symbol, atom_number):
            self.atom_counts_poscar[symbol] = number

    def read_cond(self):
        for cond_i in self.data.cond:
            for cond_file in cond_i:
                with open(cond_file, "r") as f:
                    cond = json.load(f)
                    
                # check lattice parameters
                lattice = np.array(cond["lattice"], dtype=np.float64)
                eps = 1e-3
                var = np.linalg.norm(self.lattice_poscar - lattice)
                if var > eps:
                    print(f"Error: unmatched lattice ({cond_file})")
                    sys.exit(0)
                    
                # check symbol
                if set(cond["atom_counts"].keys()) != set(self.atom_counts_poscar.keys()):
                    print(f"Error: unmatched atom species ({cond_file})")
                
                # get vacancy info
                if self.symbol_vac is None:
                    for symbol in cond["atom_counts"].keys():
                        if cond["atom_counts"][symbol] != self.atom_counts_poscar[symbol]:
                            self.symbol_vac = symbol
                            self.number_vac = self.atom_counts_poscar[symbol] - cond["atom_counts"][symbol]
                else:
                    num_vac = self.atom_counts_poscar[symbol] - cond["atom_counts"][symbol]
                    if num_vac != self.number_vac:
                        print(f"Error: unmatched number of vacancies ({cond_file})")
                        sys.exit(0)
                        
                        
def Automation_serial(data, 
                      lattice, 
                      interval,
                      num_vac,
                      tolerance: float = 1e-3,
                      use_incomplete_encounter: bool = True):
    """
    Arguments
    ---------
    data : DataInfo
        DataInfo object
    lattice : Lattice
        Lattice object
    interval : float
        Time interval (ps)
    num_vac : int
        Number of vacancies
    tolerance : flaot, optional
        Tolerance for numerical accuracy
    use_incomplete_encounter : bool, optional
        If true, incomplete encounters are used together in computations.
    """
        
    results = []
    failure = []
    task_size = len(data.datainfo)
    
    for i in tqdm(range(task_size),
                  bar_format='{l_bar}%s{bar:35}%s{r_bar}{bar:-10b}'%(Fore.GREEN, Fore.RESET),
                  ascii=False,
                  desc=f'{RED}{BOLD}Progress{RESET}'):
        
        temp, label = data.datainfo[i]
        try:
            result = Calculator(
                data=data,
                temp=temp,
                label=label,
                lattice=lattice,
                interval=interval,
                num_vac=num_vac,
                tolerance=tolerance,
                use_incomplete_encounter=use_incomplete_encounter
            )
            
        except SystemExit:
            result = Calculator_fail(
                data=data,
                temp=temp,
                label=label
            )
        
        if result.success:
            results.append(result)
            
        else:
            failure.append(
                f"  T={result.temp}K,  Label={result.label} ({result.fail_reason})"
            )
            
    # sort by (temp, label)   
    index = [data.datainfo.index([result.temp, result.label]) for result in results]
    results = [x for _, x in sorted(zip(index, results))]
    
    # print failed calculations
    if len(failure) > 0:
        print(f"Error occured in :")
        for x in failure:
            print(x)
    print('')
    
    return results


def Automation_parallel(data, 
                        lattice, 
                        interval,
                        num_vac,
                        tolerance: float = 1e-3,
                        use_incomplete_encounter: bool = True):
    """
    Arguments
    ---------
    data : DataInfo
        DataInfo object
    lattice : Lattice
        Lattice object
    interval : float
        Time interval (ps)
    num_vac : int
        Number of vacancies
    tolerance : flaot, optional
        Tolerance for numerical accuracy
    use_incomplete_encounter : bool, optional
        If true, incomplete encounters are used together in computations.
    """
    
    time_i = time.time()
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    task_size = len(data.datainfo)
    
    if rank==0:
        task_queue = list(range(task_size))
        print(f"Number of AIMD data : {len(task_queue)}")
        
        results, failure = [], []
        completed_task, terminated_worker, active_workers = 0, 0, size - 1

        while completed_task < task_size or terminated_worker < active_workers:
            status = MPI.Status()
            worker_id, task_result = comm.recv(
                source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status
            )
            
            if status.Get_tag() == 4:
                terminated_worker += 1
                continue
                
            if task_result is not None:
                completed_task += 1
                
                if task_result.success:
                    results.append(task_result)
                    state = 'success'
                    
                else:
                    failure.append(
                        f"  T={task_result.temp}K,  Label={task_result.label} ({task_result.fail_reason})"
                    )
                    state = 'fail'
                    
                print(f"Progress: {completed_task}/{task_size} finished ({state}), " +
                      f"T={task_result.temp}K Label={task_result.label}, " + 
                      f"remaining workers = {active_workers - terminated_worker}/{active_workers}")
                
            if task_queue:
                new_task = task_queue.pop()
                comm.send(new_task, dest=worker_id, tag=1)
                
            else:
                comm.send(None, dest=worker_id, tag=0)
                
        while terminated_worker < active_workers:
            worker_id, _ = comm.recv(source=MPI.ANY_SOURCE, tag=4)
            terminated_worker += 1

    else:
        comm.send((rank, None), dest=0, tag=2)
        while True:
            task = comm.recv(source=0, tag=MPI.ANY_TAG)
            
            if task is None:
                comm.send((rank, None), dest=0, tag=4)
                break
            
            try:
                temp, label = data.datainfo[task]
                result = Calculator(
                    data=data,
                    temp=temp,
                    label=label,
                    lattice=lattice,
                    interval=interval,
                    num_vac=num_vac,
                    tolerance=tolerance,
                    use_incomplete_encounter=use_incomplete_encounter
                )
                
            except SystemExit:
                result = Calculator_fail(
                    data=data,
                    temp=temp,
                    label=label
                )
                
            finally:
                comm.send((rank, result), dest=0, tag=3)
            
    if rank==0:
        index = [data.datainfo.index([result.temp, result.label]) for result in results]
        results = [x for _, x in sorted(zip(index, results))]
        
        time_f = time.time()
        
        if failure:
            print(f"\nError occured in :")
            for x in failure:
                print(x)
        print('')
        print(f"Total time taken: {time_f - time_i} s")
        
        return results
    
    
# class DataInfo:
#     def __init__(self, 
#                  prefix1: str ='traj',
#                  prefix2: str ='traj',
#                  prefix_pos: str = 'pos',
#                  prefix_force: str = 'force',
#                  prefix_cond: str = 'cond',
#                  verbose: bool =False):
        
#         self.prefix1 = prefix1
#         self.prefix2 = prefix2
#         self.prefix_pos = prefix_pos
#         self.prefix_force = prefix_force
#         self.prefix_cond = prefix_cond
#         self.verbose = verbose
        
#         # subdirectories
#         self.sub_directory = []
#         self.get_sub_directory()
        
#         self.label = []
#         self.temp = []
#         self.get_label()
#         self.check_files()
#         self.get_temperautre()
        
#         self.datainfo = [
#             [temp, label] for i, temp in enumerate(self.temp)
#             for label in self.label[i]
#         ]
#         if verbose:
#             with open('data.txt', 'w') as f:
#                 original_stdout = sys.stdout
#                 sys.stdout = f
#                 try:
#                     self.summary()
#                 finally:
#                     sys.stdout = original_stdout
    
#     def get_temperautre(self):
#         self.temp = []
#         for i, dir in enumerate(self.sub_directory):
#             cond_file = os.path.join(
#                 self.prefix1, dir, f"cond_{self.label[i][0]}.json"
#             )
#             with open(cond_file, "r") as f:
#                 condition = json.load(f)
#             self.temp.append(condition['temperature'])
            
#     def check_files(self):
#         check = False
#         for i, dir in enumerate(self.sub_directory):
#             for label in self.label[i]:
#                 # check cond.json
#                 cond_file = os.path.join(
#                     self.prefix1, dir, f"cond_{label}.json"
#                 )
#                 pos_file = os.path.join(
#                     self.prefix1, dir, f"pos_{label}.npy"
#                 )
#                 force_file = os.path.join(
#                     self.prefix1, dir, f"force_{label}.npy"
#                 )
#                 if not os.path.isfile(cond_file):
#                     print(f"Error: {cond_file} is not found.")
#                     check = True
#                 if not os.path.isfile(pos_file):
#                     print(f"Error: {pos_file} is not found.")
#                     check = True
#                 if not os.path.isfile(force_file):
#                     print(f"Error: {force_file} is not found.")
#                     check = True
#         if check:
#             sys.exit(0)
        
#     def get_sub_directory(self):
#         list_dir = os.listdir(os.path.join(os.getcwd(), self.prefix1))
#         self.sub_directory = [d for d in list_dir if self.prefix2 in d]
#         self.sub_directory.sort()   
    
#     def get_label(self):
#         for dir in self.sub_directory:
#             list_file = os.listdir(
#                 os.path.join(self.prefix1, dir)
#             )
#             list_file.sort()
            
#             label = []
#             for file in list_file:
#                 extension = file.split(".")[-1]
#                 if extension == "npy":
#                     _file = "".join(file.split(".")[:-1]).split('_')
#                     if _file[0] == self.prefix_pos and len(_file) > 1:
#                         label.append(_file[1])
#             self.label.append(label)
    
#     def summary(self):
#         label_all = list(
#             {element for sublabel in self.label for element in sublabel}
#         )
#         label_all.sort()
        
#         print("# List of data :")
#         header = ['label'] + [f"{temp}K" for temp in self.temp]
#         data = [
#             [label] + ['O' if label in self.label[j] else 'X' for j in range(len(self.temp))]
#             for label in label_all
#         ]
#         data.append(
#             ['Total'] + [len(label) for label in self.label]
#         )
#         print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
#         print('')
             

# class VacancyInfo:
#     def __init__(self,
#                  data,
#                  poscar_lattice: str):
#         """
#         Arguments
#         ---------
#         data : Data
#             Data object
#         poscar_lattice : str
#             Path to POSCAR_LATTICE (POSCAR of pefect crystalline)
#         """
        
#         self.data = data
#         self.poscar_lattice = poscar_lattice
        
#         self.atom_symbol_poscar = None
#         self.atom_number_poscar = None
#         self.read_poscar()
        
#         self.symbol_vac = None
#         self.number_vac = None
#         self.read_xdatcar()
        
#     def read_poscar(self):
#         # perfect crystal structure
#         check_symbol, check_number = False, False
#         with open(self.poscar_lattice, 'r') as f:
#             for i, line in enumerate(f):
#                 if i == 5:
#                     self.atom_symbol_poscar = line.split()
#                     check_symbol = True
                    
#                 if i == 6:
#                     self.atom_number_poscar = list(map(int, line.split()))
#                     check_number = True
                    
#                 if check_symbol and check_number:
#                     break
                
#     def read_xdatcar(self):
#         xdatcar = self.data.xdatcar
        
#         atom_symbol_xdatcar = []
#         atom_number_xdatcar = []
#         _atom_number_xdatcar = None
        
#         for xdatcar_temp in xdatcar:
#             for xdatcar_i in xdatcar_temp:
#                 check_symbol, check_number = False, False
#                 with open(xdatcar_i, 'r') as f:
#                     for i, line in enumerate(f):
#                         if i == 5:
#                             symbol = line.split()
#                             if symbol != self.atom_symbol_poscar:
#                                 print(f"Error: unmatched atom species ({xdatcar_i}).")
#                                 sys.exit(0)
#                             atom_symbol_xdatcar.append(line.split())
#                             check_symbol = True
                            
#                         if i == 6:
#                             number = list(map(int, line.split()))
#                             if _atom_number_xdatcar is None:
#                                 _atom_number_xdatcar = number
                                
#                                 for j, (n1, n2) in enumerate(
#                                     zip(self.atom_number_poscar, number)
#                                     ):
#                                     if n1 != n2:
#                                         self.number_vac = n1 - n2
#                                         self.symbol_vac = self.atom_symbol_poscar[j]
#                                         break
#                             else:
#                                 if number != _atom_number_xdatcar:
#                                     print(f"Error: unmatched atom numbers ({xdatcar_i}).")
#                                     sys.exit(0)
#                                 atom_number_xdatcar.append(number)
                
#                             check_number = True
                            
#                         if check_symbol and check_number:
#                             break   


import os
import sys
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from tabulate import tabulate
from scipy.optimize import minimize_scalar
from vachoppy.trajectory import *


class Calculator:
    def __init__(self,
                 data,
                 temp,
                 label,
                 lattice,
                 interval,
                 num_vac,
                 tolerance: float = 1e-3,
                 use_incomplete_encounter: bool = True):
        """
        Arguments
        ---------
        data : DataInfo
            DataInfo object
        temp : float
            Temperature (K)
        label : str
            Label
        lattice : Lattice
            Lattice object
        interval : float
            Time interval (ps)
        num_vac : int
            Number of vacancies
        tolerance : flaot, optional
            Tolerance for numerical accuracy
        use_incomplete_encounter : bool, optional
            If true, incomplete encounters are used together in computations.
        """
        
        # arguments
        self.data = data
        self.temp = temp
        self.label = label
        self.lattice = lattice
        self.interval = interval
        self.num_vac = num_vac
        self.tolerance = tolerance
        self.use_incomplete_encounter = use_incomplete_encounter
        
        self.index_temp = None
        self.index_label = None
        self.pos = None
        self.cond = None
        self.force = None
        self.get_address()
        
        # quantities from trajectory
        self.potim = None
        self.num_step = None
        self.transient_vacancy = None
                
        # quantities from analyzer
        self.num_path = len(lattice.path_name)
        self.hopping_history = None
        self.counts = None
        self.path_unknown = None
        self.unknown_name = None
        self.counts_unknown = None
        self.residence_time = None
        self.msd_rand = None
        
        # quantities from encounter
        self.f_cor = None
        self.encounter_num = None
        self.encounter_msd = None
        self.encounter_path_names = None
        self.encounter_path_counts = None
        self.encounter_path_distance = None
        self.extract_AIMD_quantities()
        
        # check success
        self.success = True
        self.fail_reason = None
        
    def get_address(self):
        for i, temp_i in enumerate(self.data.temp):
            if abs(temp_i - self.temp) < self.tolerance:
                self.index_temp = i
                break
        
        for i, label_i in enumerate(self.data.label[self.index_temp]):
            if label_i == self.label:
                self.index_label = i
                break
            
        if None in [self.index_temp, self.index_label]:
            print(f"No data exists (T : {self.temp} K, Label : {self.label})")
            sys.exit(0)
            
        self.pos = self.data.pos[self.index_temp][self.index_label]
        self.cond = self.data.cond[self.index_temp][self.index_label]
        self.force = self.data.force[self.index_temp][self.index_label]
         
    def extract_AIMD_quantities(self):
        # trajectory
        try:
            trajectory = Trajectory(
                interval=self.interval,
                num_vac=self.num_vac,
                lattice=self.lattice,
                pos_file=self.pos,
                force_file=self.force,
                cond_file=self.cond,
                verbose=False
            )
        
        except SystemExit:
            self.success = False
            self.fail_reason = "Error by trajectory.Trajectory"
            return
        
        except BaseException as e:
            self.success = False
            self.fail_reason = "Error by trajectory.Trajectory"
            print(e)
            return
        
        self.potim = trajectory.potim
        self.num_step = trajectory.num_step
        self.transient_vacancy = trajectory.transient_vacancy

        # analyzer
        try:
            analyzer = TrajectoryAnalyzer(
                lattice=self.lattice,
                trajectory=trajectory,
                tolerance=self.tolerance,
                verbose=False
            )
        except SystemExit:
            self.success = False
            self.fail_reason = "Error by trajectory.TrajectoryAnalyzer"
            return
        
        except BaseException as e:
            self.success = False
            self.fail_reason = "Error by trajectory.TrajectoryAnalyzer"
            print(e)
            return
        
        self.hopping_history = analyzer.hopping_history
        self.counts = analyzer.counts
        self.path_unknown = analyzer.path_unknown
        self.unknown_name = analyzer.unknown_name
        self.counts_unknown = analyzer.counts_unknown
        self.residence_time = analyzer.residence_time
        self.msd_rand = analyzer.msd_rand
        
        # encounter
        try:
            encounter = Encounter(
                analyzer=analyzer,
                use_incomplete_encounter=self.use_incomplete_encounter,
                verbose=False
            )
        except SystemExit:
            self.success = False
            self.fail_reason = "Error by trajectory.TrajectoryAnalyzer"
            return
        
        except BaseException as e:
            self.success = False
            self.fail_reason = "Error by trajectory.Encounter"
            print(e)
            return
        
        self.f_cor = encounter.f_cor
        self.encounter_num = encounter.num_encounter
        self.encounter_msd = encounter.msd
        self.encounter_path_names = encounter.path_name
        self.encounter_path_counts = encounter.path_count
        self.encounter_path_distance = encounter.path_distance

        
class Calculator_fail:
    def __init__(self, 
                 data, 
                 temp,
                 label):
        self.data = data
        self.temp = temp
        self.label = label
        self.success = False
        self.fail_reason = 'Unknown reason'


class ParameterExtractor:
    def __init__(self,
                 data,
                 results : list,
                 tolerance: float = 1e-3,
                 verbose: bool = True,
                 figure: bool = True,
                 file_out: str = 'parameter.txt',
                 inset_correlatoin_factor: bool = True):
        """
        Arguments
        ---------
        data : DataInfo
            DataInfo object
        results : list
            List of Calculator objects
        tolerance : float, optional
            Numerial tolerance
        verbose : bool, optional
            Verbosity flag
        figure : bool, optional
            If True, figures will generated
        file_out : str, optional
            File name which the computational summary will be saved in
        inset_correlatoin_factor : bool, optional
            If True, Arrhenius plot for correlation factor is shown together
        """
        # arguments
        self.data = data
        self.results = results
        self.lattice = results[-1].lattice
        self.tolerance = tolerance
        self.verbose = verbose
        self.figure = figure
        
        self.num_vac = results[-1].num_vac
        self.temp = np.array(self.data.temp, dtype=np.float64)
        self.num_temp = len(self.temp)
        
        # constants
        self.kb = 8.61733326e-5
        self.cmap = plt.get_cmap("Set1")
        
        # unknown paths
        self.path_unknown = self.lattice.path_unknown
        
        # labels for successful results
        self.num_label = [
            [result.temp for result in self.results].count(temp) 
            for temp in self.temp
        ]
        self.index = np.concatenate(([0], np.cumsum(self.num_label)))
        
        # correlation factor
        self.f_ind = None
        self.f_avg = None
        self.f_cum = None
        self.correlation_factor()
        
        # diffusion coefficient
        self.D_rand = None
        self.D0_rand = None
        self.Ea = None
        self.random_walk_diffusion_coefficient()
        
        # residence time
        self.tau = None
        self.tau0 = None
        self.residence_time()
        
        # hopping distance
        self.a_eff = np.sqrt(6 * self.D0_rand * self.tau0) * 1e4
        
        # <z>
        self.z_mean = None
        self.mean_number_of_equivalent_paths()
        
        if verbose:
            if file_out is None:
                self.summary()
            else:
                with open(file_out, 'w') as f:
                    original_stdout = sys.stdout
                    sys.stdout = f
                    try:
                        self.summary()
                    finally:
                        sys.stdout = original_stdout
        
        if figure:
            self.save_figure(inset_correlatoin_factor=inset_correlatoin_factor)  
        
    def correlation_factor(self):
        self.f_ind = []
        self.f_avg = []
        self.f_cum = []
        
        # individual correlatoin factor
        self.f_ind = np.array([result.f_cor for result in self.results], dtype=np.float64)
        
        # average at each temperature
        self.f_avg = [
            np.mean([f for f in self.f_ind[start:end] if not np.isnan(f)])
            for start, end in zip(self.index[:-1], self.index[1:])
        ]
        
        # cumulative correlatoin factor
        for i in range(self.num_temp):
            num_encounter = []
            msd_encounter = []
            msd_encounter_rand = []
            
            for j in range(self.index[i], self.index[i+1]):
                
                if self.results[j].f_cor is None:
                    continue
                
                num_encounter.append(self.results[j].encounter_num)
                msd_encounter.append(self.results[j].encounter_msd)
                msd_encounter_rand.append(
                    np.sum(
                        self.results[j].encounter_path_distance**2 * self.results[j].encounter_path_counts
                    )
                )
                
            msd_encounter = np.array(msd_encounter)
            num_encounter = np.array(num_encounter)
            msd_encounter_rand = np.array(msd_encounter_rand)

            msd_encounter = np.sum(msd_encounter * num_encounter) / np.sum(num_encounter)
            msd_encounter_rand = np.sum(msd_encounter_rand) / np.sum(num_encounter)
            
            f_cum_i = msd_encounter / msd_encounter_rand
            self.f_cum.append(f_cum_i)

    def random_walk_diffusion_coefficient(self):
        # random walk diffusion coefficient
        self.D_rand = []
        for start, end in zip(self.index[:-1], self.index[1:]):
            t_i = np.sum(
                np.array(
                    [result.interval * result.num_step for result in self.results[start:end]]
                )
            )
            msd_rand_i = np.sum(
                np.array(
                    [result.msd_rand for result in self.results[start:end]]
                )
            )
            self.D_rand.append(msd_rand_i / (6 * t_i))
        self.D_rand = np.array(self.D_rand) * 1e-8
        
        # Arrhenius fit
        slop, intercept = np.polyfit(1/self.temp, np.log(self.D_rand), deg=1)
        self.Ea = -slop * self.kb
        self.D0_rand = np.exp(intercept)
        
    def residence_time(self):
        self.tau = []
        for start, end in zip(self.index[:-1], self.index[1:]):
            t_i = np.sum(
                np.array(
                    [result.interval * result.num_step for result in self.results[start:end]]
                )
            )
            count_i = np.sum(
                [np.sum(result.counts) / result.num_vac for result in self.results[start:end]]
            )
            self.tau.append(t_i / count_i)
        self.tau = np.array(self.tau)
        
        # Arrhenius fit
        error_tau = lambda tau0: np.linalg.norm(
            self.tau - tau0 * np.exp(self.Ea / (self.kb * self.temp))
        )
        tau0_opt = minimize_scalar(error_tau)
        self.tau0 = tau0_opt.x # ps
            
    def mean_number_of_equivalent_paths(self):
        self.z_mean = []
        z = np.array(
            [path['z'] for path in self.lattice.path], dtype=np.float64
        )
        
        for start, end in zip(self.index[:-1], self.index[1:]):
            counts = np.array(
                [np.sum(result.counts, axis=0) for result in self.results[start:end]]
            )
            counts = np.sum(counts, axis=0)
            
            self.z_mean.append(np.sum(counts) / np.sum(counts / z))
            
    def print_lattice_info(self):
        path_site = [path['site_init'] for path in self.lattice.path]
        num_path_site = [path_site.count(name) for name in self.lattice.site_name]
        
        print('Lattice information :')
        print('  Vacancy type : ', self.lattice.symbol)
        print(f"  Number of vacancies : {self.num_vac}")
        print('  Number of sites : ', len(self.lattice.site_name))
        print('  Number of hopping paths : ', end='')
        for num in num_path_site:
            print(int(num), end=' ')
        print('\n  Number of unknown paths =', len(self.path_unknown))
        print('')
        
        print('Vacancy hopping paths : ')
        header = ['path', 'a (Å)', 'z', 'initial site', 'final site']
        data = [
            [path['name'],  
             f"{path['distance']:.3f}", path['z'], 
             path['site_init'] + f" [{path['coord_init'][0]:.5f} {path['coord_init'][1]:.5f} {path['coord_init'][2]:.5f}]",
             path['site_final'] + f" [{path['coord_final'][0]:.5f} {path['coord_final'][1]:.5f} {path['coord_final'][2]:.5f}]"]
            for path in self.lattice.path
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        print('Unknown hopping paths : ')
        header = ['path', 'a (Å)', 'z', 'initial site', 'final site']
        data = [
            [path['name'],  
             f"{path['distance']:.3f}", '-', 
             path['site_init'] + f" [{path['coord_init'][0]:.5f} {path['coord_init'][1]:.5f} {path['coord_init'][2]:.5f}]",
             path['site_final'] + f" [{path['coord_final'][0]:.5f} {path['coord_final'][1]:.5f} {path['coord_final'][2]:.5f}]"]
            for path in self.path_unknown
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
    def print_simulation_condition(self):
        print(f'Time ineterval used for average (t_interval) : {self.results[0].interval} ps')
        print('')
        
        print('Simulation temperatures (K) : ', end='')
        for temp in self.temp:
            print(temp, end=' ')
        print('\n')   
        
        print('Labels of AIMD data :')
        label_all = list(
            {element for sublabel in self.data.label for element in sublabel}
        )
        label_all.sort()
        
        header = ['label'] + [f"{temp}K" for temp in self.temp]
        data = [
            [label] + ['O' if label in self.data.label[j] else 'X' for j in range(len(self.temp))]
            for label in label_all
        ]
        data.append(['Num'] + self.num_label)
        
        print(tabulate(data, headers=header, tablefmt="simple", stralign='center', numalign='center'))
        print('')
        
    def print_effective_hopping_parameter(self):
        print('Effective hopping parameters : ')
        header = ['parameter', 'value', 'description']
        parameter = [
            'Drand_0 (m2/s)', 'tau0 (ps)', 'Ea (eV)', 'a (Å)', 'f', '<z>'
        ]
        value = [
            f"{self.D0_rand:.5e}", 
            f"{self.tau0:.5f}", 
            f"{self.Ea:.5f}", 
            f"{self.a_eff:.5f}",
            f"{np.average(self.f_cum):.5f}", 
            f"{np.average(self.z_mean):.5f}"
        ]
        description = [
            'pre-exponential for random walk diffusivity',
            'pre-exponential for residence time',
            'hopping barrier',
            'hopping distance',
            'correlation factor',
            'mean number of equivalent paths per path type'
        ]
        data = [[p, v, d] for p, v, d in zip(parameter, value, description)]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')   

    def print_diffusion_coefficient(self):
        print('Random walk diffusion coefficient : ')
        header = ['T (K)', 'D_rand (m2/s)']
        data = [
            [f"{temp}", f"{D:.5e}"] for temp, D in zip(self.temp, self.D_rand)
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
    
    def print_residence_time(self):
        print('Residence time : ')
        header = ['T (K)', 'tau (ps)']
        data = [
            [f"{temp}", f"{tau:.5f}"] for temp, tau in zip(self.temp, self.tau)
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
    
    def print_correlation_factor(self):
        label_all = list(
            {element for sublabel in self.data.label for element in sublabel}
        )
        label_all.sort()
        
        # rearrange f_ind
        f_ind_re = []
        index = [0] + list(np.cumsum(self.num_label))
        for i, f in enumerate(range(len(self.temp))):
            f_ind_re_i = ['-'] * len(label_all)
            for j, cal in enumerate(self.results[index[i]:index[i+1]]):
                f_ind_re_i[label_all.index(cal.label)] = f"{self.f_ind[index[i]+j]:.5f}"
            f_ind_re.append(f_ind_re_i)
        f_ind_re = [list(x) for x in zip(*f_ind_re)] # transpose
        
        # check use_incomplete_encounter
        if self.results[-1].use_incomplete_encounter:
            print("Caution! use_incomplete_encounter = True was set.")
            print("Incompleted encounters are also used for correlatoin factor calculations")
            print("use_incomplete_encounter = True is recommended only when simulation time is not enough to catch sufficient encounters")
            print('')
            
        print('Cumulative correlation factors : ')
        print('(Note: use these values for your work)')
        header = ['T (K)', 'f']
        data = [
            [f"{temp}", f"{f:.5f}"] for temp, f in zip(self.temp, self.f_cum)
        ]
        data.append(['Average', f"{np.average(self.f_cum):.5f}"])
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        print("Individual correlation factors :")
        print('(Note: use these values only for convergence tests)')
        header = ['label'] + [f"f({int(T)}K)" for T in self.temp]
        data = [
            [label] + f_ind_re_i for label, f_ind_re_i in zip(label_all, f_ind_re)
        ]
        data.append(
            ['Average'] + [f"{f:.5f}" for f in self.f_avg]
        )
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
    
    def print_counts(self):
        name = self.lattice.path_name + self.results[-1].unknown_name
        counts = []
        for start, end in zip(self.index[:-1], self.index[1:]):
            num_normal = len(self.lattice.path)
            counts_i = np.zeros((self.num_vac, len(name)))
            
            for result in self.results[start:end]:
                # normal path
                counts_i[:, :num_normal] += result.counts
                
                # unknown path
                num_unknown = result.counts_unknown.shape[-1]
                if num_unknown > 0:
                    counts_i[:, num_normal:num_normal+num_unknown] += result.counts_unknown       
            
            counts_i = np.sum(counts_i, axis=0)
            counts.append(list(counts_i))          
                    
        print("Counts for each hopping path :")
        header = ['T (K)'] + name
        data = [
            [temp] + counts_i for temp, counts_i in zip(self.temp, counts)
        ]
        counts = np.array(counts)
        data.append(['Total'] + np.sum(counts, axis=0).tolist())
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
    
    def print_total_residence_time(self):
        t_simul = []
        t_reside = []
        for start, end in zip(self.index[:-1], self.index[1:]):
            t_simul_i = 0
            t_reside_i = np.zeros_like(self.lattice.site_name, dtype=np.float64)
            
            for result in self.results[start:end]:
                t_simul_i += result.num_step * result.interval * result.num_vac
                t_reside_i += np.sum(result.residence_time, axis=0)
            t_simul.append(t_simul_i)
            t_reside.append(t_reside_i)
        t_simul = np.array(t_simul)
        t_reside = np.array(t_reside)       
        
        print('Time vacancy remained at each site (ps) :')
        header = ['T (K)'] + self.lattice.site_name + ['Total']
        data = [
            [temp] + t_reside_i.tolist() + [t_simul_i]
            for temp, t_reside_i, t_simul_i in zip(self.temp, t_reside, t_simul)
        ]
        data.append(['Total'] + np.sum(t_reside, axis=0).tolist() + [np.sum(t_simul)])
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
    
    def print_lattice_point(self):
        site_num = []
        site_type = []
        site_coord = []
        for i, site in enumerate(self.lattice.lattice_point):
            site_num.append(i)
            site_type.append(site['site'])
            site_coord.append(site['coord'])
        
        print('Lattice point information : ')
        header = ['name', 'site_type', 'frac_coord']
        data = [
            [f"{self.lattice.symbol}{num}", t, f"[{c[0]:.5f} {c[1]:.5f} {c[2]:.5f}]"]
            for num, t, c in zip(site_num, site_type, site_coord)
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
    
    def print_hopping_history(self):
        print("Computational details :")
        for result in self.results:
            print(f"-----------------(T = {result.temp} K Label = {result.label})-----------------")
            print("Vacancy hopping history :")
            for i in range(self.num_vac):
                print(f"# Vacancy{i+1}")
                header = ['num', 'time (ps)', 'path', 'a (Å)', 'initial site', 'final site']
                data = [
                    [
                        f"{j+1}",
                        f"{path['step'] * result.interval:.2f}",
                        f"{path['name']}",
                        f"{path['distance']:.5f}",
                        f"{path['site_init']} [{', '.join(f'{x:.5f}' for x in self.lattice.lattice_point[path['index_init']]['coord'])}]",
                        f"{path['site_final']} [{', '.join(f'{x:.5f}' for x in self.lattice.lattice_point[path['index_final']]['coord'])}]"
                    ] for j, path in enumerate(result.hopping_history[i])
                ]
                print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
                print('')
            
            print("Summary :")
            
            name_all = self.lattice.path_name + result.unknown_name
            counts_all = np.hstack((result.counts, result.counts_unknown))
            counts_all = np.array(counts_all, dtype=np.int32)
            vacancy_name = [f"Vacancy{i+1}" for i in range(result.num_vac)]
            print("# Path counts :")
            header = ['path'] + vacancy_name
            data = np.vstack((name_all, counts_all)).T
            print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
            print('')
            
            print("# Transient vacancies :")
            step_transient = [k for k, d in result.transient_vacancy.items() if len(d) > 0]
            header = ['step', 'number of transient vacancies', 'site indidces']
            data = [
                [str(k), 
                 len(result.transient_vacancy[k]),
                 " ".join(map(str, result.transient_vacancy[k]))
                ] for k in step_transient
            ]
            print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
            print('') 
            
            print("# Vacancy residence time (ps) :")
            header = ['site'] + vacancy_name
            data = np.vstack((self.lattice.site_name, result.residence_time)).T
            print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
            print('')
    
    def save_figure(self,
                    inset_correlatoin_factor=True):
        # D_rand
        plt.style.use('default')
        plt.rcParams['figure.figsize'] = (3.8, 3.8)
        plt.rcParams['font.size'] = 11
        fig, ax = plt.subplots()
        for axis in ['top','bottom','left','right']:
            ax.spines[axis].set_linewidth(1.2)
        
        for i in range(len(self.temp)):
            plt.scatter(1/self.temp[i], np.log(self.D_rand[i]), 
                        color=self.cmap(i), marker='s', s=50, label=str(int(self.temp[i])))
        slop, intercept = np.polyfit(1/self.temp, np.log(self.D_rand), deg=1)
        x = np.linspace(np.min(1/self.temp), np.max(1/self.temp), 100)
        plt.plot(x, slop*x + intercept, 'k:', linewidth=1)
        plt.xlabel('1/T (1/K)', fontsize=14)
        plt.ylabel(r'ln $D_{rand}$ ($m^{2}$/s)', fontsize=14)
        num_data = len(self.D_rand)
        ncol = int(np.ceil(num_data / 5))
        plt.legend(loc='best', fancybox=True, framealpha=1, edgecolor='inherit',
                   ncol=ncol, labelspacing = 0.3, columnspacing=0.5, borderpad=0.2, handlelength=0.6,
                   fontsize=11, title='T (K)', title_fontsize=11)
        if num_data >= 3:
            x = np.array([self.temp[0], self.temp[int(num_data/2)], self.temp[-1]])
        else:
            x = self.temp
        x_str = [f"1/{int(T)}" for T in x]
        x = 1/x
        plt.xticks(x, x_str)
        plt.savefig('D_rand.png', transparent=False, dpi=300, bbox_inches="tight")
        plt.close()
        
        # tau
        plt.style.use('default')
        plt.rcParams['figure.figsize'] = (3.8, 3.8)
        plt.rcParams['font.size'] = 11
        fig, ax = plt.subplots()
        for axis in ['top','bottom','left','right']:
            ax.spines[axis].set_linewidth(1.2)
            
        for i, temp in enumerate(self.temp):
            ax.bar(temp, self.tau[i], width=50, edgecolor='k', color=self.cmap(i))
            ax.scatter(temp, self.tau[i], marker='o', edgecolors='k', color='k')
        x = np.linspace(0.99*self.temp[0], 1.01*self.temp[-1], 1000)
        ax.plot(x, self.tau0 * np.exp(self.Ea/(self.kb*x)), 'k:')
        plt.xlabel('T (K)', fontsize=14)
        plt.ylabel(r'$\tau$ (ps)', fontsize=14)
        plt.savefig('tau.png', transparent=False, dpi=300, bbox_inches="tight")
        plt.close()
        
        # correlation factor
        plt.style.use('default')
        plt.rcParams['figure.figsize'] = (3.8, 3.8)
        plt.rcParams['font.size'] = 11
        fig, ax = plt.subplots()
        for axis in ['top','bottom','left','right']:
            ax.spines[axis].set_linewidth(1.2)
            
        for i in range(len(self.temp)):
            ax.scatter(self.temp[i], self.f_cum[i], color=self.cmap(i), marker='s', s=50)
        plt.ylim([0, 1])
        plt.xlabel('T (K)', fontsize=14)
        plt.ylabel(r'$f$', fontsize=14)
        
        if inset_correlatoin_factor:
            slop, intercept = np.polyfit(1/self.temp, np.log(self.f_cum), deg=1)
            Ea_f = -slop * self.kb
            f0 = np.exp(intercept)
            
            # inset graph
            axins = ax.inset_axes([1.125, 0.615, 0.35, 0.35])
            x_ins = np.linspace(1/self.temp[-1], 1/self.temp[0], 100)
            axins.plot(x_ins, slop * x_ins + intercept, 'k:')
            for i in range(len(self.temp)):
                axins.scatter(1/self.temp[i], np.log(self.f_cum[i]), color=self.cmap(i), marker='s')
            axins.set_xlabel('1/T', fontsize=12)
            axins.set_ylabel(r'ln $f$', fontsize=12)
            axins.set_xticks([])
            axins.set_yticks([])
            
            ax.text(1.125, 0.4, 
                    r"$E_{\mathrm{a}} = $" + f"{Ea_f:.3f} eV", 
                    transform=ax.transAxes, fontsize=10)
            ax.text(1.125, 0.3, 
                    r"$f_{\mathrm{0}} = $" + f"{f0:.3f}", 
                    transform=ax.transAxes, fontsize=10)

        plt.savefig('f_cor.png', transparent=False, dpi=300, bbox_inches="tight")
        plt.close()
        print('')    
        
    def summary(self):
        self.print_lattice_info()
        self.print_simulation_condition()
        self.print_effective_hopping_parameter()
        self.print_diffusion_coefficient()
        self.print_residence_time()
        self.print_correlation_factor()
        self.print_counts()
        self.print_total_residence_time()
        self.print_lattice_point()
        self.print_hopping_history()

        
class PostProcess:
    def __init__(self, 
                 file_params='parameter.txt',
                 file_neb = 'neb.csv',
                 verbose=False):
        # check file
        if os.path.isfile(file_params):
            self.file_params = file_params
        else:
            print(f"{file_params} is not found.")
            sys.exit(0)
        if os.path.isfile(file_neb):
            self.file_neb = file_neb
        else:
            print(f"{file_neb} is not found.")
            sys.exit(0)
        self.verbose = verbose
        self.kb = 8.61733326e-5
        
        # read parameter file
        self.num_sites = None
        self.num_paths = None
        self.path_names = []
        self.z = []
        self.temp = None
        self.times = []
        self.counts = []
        self.D0_eff = None
        self.Ea_eff = None
        self.tau0_eff = None
        self.a_eff = None
        self.f = []
        self.f_mean = None
        self.read_parameter()
        
        # read neb file
        self.Ea = None
        self.read_neb()
        
        # P_site
        self.P_site = self.times / np.sum(self.times, axis=1).reshape(-1,1)
        
        # P_esc
        self.P_esc = np.exp(-self.Ea/(self.kb * self.temp[:, np.newaxis]))
        self.P_esc_eff = np.exp(-self.Ea_eff / (self.kb * self.temp))
        
        # P = P_site * P_esc
        self.P = None
        self.get_P()
        
        # z_mean
        self.z_mean = None
        self.z_mean_rep = None # from total counts from all temperatures
        self.get_z_mean()
        
        # z_eff
        self.z_eff = np.sum(self.P * self.z, axis=1) / self.P_esc_eff
        self.z_eff_rep = np.average(self.z_eff)
        
        # <m>
        self.m_mean = self.z_eff / self.z_mean
        self.m_mean_rep = np.average(self.m_mean)
        
        # nu
        self.nu = None
        self.nu_eff = None
        self.nu_eff_rep = None # simple average of nu_eff
        self.get_nu()
        
        if self.verbose:
            self.summary()
        
    def read_parameter(self):
        with open(self.file_params, 'r') as f:
            lines = [line.strip() for line in f]
            
        for i, line in enumerate(lines):
            if "Lattice information :" in line:
                self.num_sites = int(lines[i+3].split()[-1])
                self.num_paths = list(map(int, lines[i+4].split()[-self.num_sites:]))
                self.num_paths = np.array(self.num_paths)
                
            if "Vacancy hopping paths :" in line:
                for j in range(np.sum(self.num_paths)):
                    contents = lines[i+j+3].split()
                    self.path_names.append(contents[0])
                    self.z.append(int(contents[2]))
                self.z = np.array(self.z, dtype=float)
                    
            if "Simulation temperatures (K) :" in line:
                self.temp = np.array(list(map(float, lines[i].split()[4:])), dtype=float)
                
            if "Time vacancy remained at each site (ps) :" in line:
                for j in range(len(self.temp)):
                    self.times.append(
                        list(map(float, lines[i+j+3].split()[1:1+self.num_sites]))
                    )
                self.times = np.array(self.times)
                
            if "Counts for each hopping path :" in line:
                for j in range(len(self.temp)):
                    self.counts.append(
                        list(map(int, lines[i+j+3].split()[1:1+np.sum(self.num_paths)]))
                    )
                self.counts = np.array(self.counts, dtype=float)
                
            if "Effective hopping parameters :" in line:
                self.D0_eff = float(lines[i+3].split()[2])
                self.tau0_eff = float(lines[i+4].split()[2])
                self.Ea_eff = float(lines[i+5].split()[2])
                self.a_eff = float(lines[i+6].split()[2])
                self.f_mean = float(lines[i+7].split()[1]) # no unit
                
            if "Cumulative correlation factors :" in line:
                self.f =[float(lines[i+j+4].split()[1]) for j in range(len(self.temp))]
    
    def read_neb(self):
        neb = pd.read_csv(self.file_neb, header=None).to_numpy()
        self.Ea = np.zeros(len(self.path_names), dtype=float)
        for name_i, Ea_i in neb:
            index = self.path_names.index(name_i)
            self.Ea[index] = float(Ea_i)
    
    def get_P(self):
        P_site_extend = []
        for p_site in self.P_site:
            P_site_i = []
            for p, m in zip(p_site, self.num_paths):
                P_site_i += [float(p)] * m
            P_site_extend.append(P_site_i)
        self.P = np.array(P_site_extend) * self.P_esc
    
    def get_z_mean(self):
        self.z_mean = np.sum(self.counts, axis=1) / np.sum(self.counts / self.z, axis=1)
        self.z_mean_rep = np.sum(self.counts) / np.sum(np.sum(self.counts, axis=0) / self.z)
        
    def get_nu(self):
        times_extend = []
        for time in self.times:
            times_i = []
            for t, m in zip(time, self.num_paths):
                times_i += [float(t)] * m
            times_extend.append(times_i)
        self.nu = self.counts / (self.z * self.P_esc * times_extend) 
        self.nu_eff = np.sum(self.counts, axis=1) / (np.sum(self.times, axis=1) * np.sum(self.P * self.z, axis=1))
        self.nu_eff_rep = np.average(self.nu_eff)
        
    def summary(self):
        # effective parameters
        print("Effective hopping parameters :")
        header = ['parameter', 'value', 'description']
        parameter = ["Drand_0 (m2/s)", 
                     "tau0 (ps)", 
                     "Ea (eV)", 
                     "a (Å)", 
                     "z", 
                     "nu (THz)", 
                     "f", 
                     "<z>", 
                     "<m>"]
        value = [f"{self.D0_eff:.5e}", 
                 f"{self.tau0_eff:.5e}", 
                 f"{self.Ea_eff:.5f}", 
                 f"{self.a_eff:.5f}", 
                 f"{self.z_eff_rep:.5f}", 
                 f"{self.nu_eff_rep:.5f}", 
                 f"{self.f_mean:.5f}", 
                 f"{self.z_mean_rep:.5f}", 
                 f"{self.m_mean_rep:.5f}"]
        description = ['pre-exponential for random walk diffusivity',
                       'pre-exponential for residence time',
                       'hopping barrier',
                       'hopping distance',
                       'number of equivalent paths (coordination number)',
                       'jump attempt frequency',
                       'correlation factor',
                       'mean number of equivalent paths per path type',
                       'mean number of path types (=z/<z>)']
        data = [[p, v, d] for p, v, d in zip(parameter, value, description)]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        # temperature dependence
        print("Effective hopping parameters with respect to temperature :")
        header = ["T (K)", 
                  "z", 
                  "nu (THz)", 
                  "f",
                  "<z>", 
                  "<m>"]
        data = [
            [int(T), 
             f"{z:.5f}", 
             f"{nu:.5f}",
             f"{f:.5f}",
             f"{z_mean:.5f}", 
             f"{m_mean:.5f}"
             ] 
            for T, z, nu, f, z_mean, m_mean in \
                zip(self.temp, self.z_eff, self.nu_eff, self.f, self.z_mean, self.m_mean) 
        ]
        data.append(['Average', 
                     f"{np.average(self.z_eff):.5f}",
                     f"{np.average(self.nu_eff):.5f}",
                     f"{np.average(self.f):.5f}",
                     f"{np.average(self.z_mean):.5f}", 
                     f"{np.average(self.m_mean):.5f}"])
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        # nu with respect to temperature
        print("Jump attempt frequency (THz) with respect to temperature :")
        print("(Note: only paths with sufficient sampling are reliable)")
        header = ["T (K)"] + self.path_names
        data = [
            [int(T)] + list(nu) for T, nu in zip(self.temp, self.nu)
        ]
        data.append(['Average'] + [f"{nu:.5f}" for nu in np.average(self.nu, axis=0)])
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        # P_site with respect to temperature
        print("P_site with respect to temperature :")
        header = ["T (K)"] + [f"site{i+1}" for i in range(self.num_sites)]
        data = [
            [int(T)] + [f"{p:.5e}" for p in p_i] for T, p_i in zip(self.temp, self.P_site)
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')
        
        # P_esc with respect to temperature
        print("P_esc with respect to temperature :")
        header = ["T (K)"] + self.path_names
        data = [
            [int(T)] + [f"{p:.5e}" for p in p_i] for T, p_i in zip(self.temp, self.P_esc)
        ]
        print(tabulate(data, headers=header, tablefmt="simple", stralign='left', numalign='left'))
        print('')